#!/usr/bin/env python3
# Copyright © 2019-2020 Intel Corporation

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Generates release notes for a given version of mesa."""

import asyncio
import datetime
import os
import pathlib
import re
import subprocess
import sys
import textwrap
import typing
import urllib.parse

import aiohttp
from mako.template import Template
from mako import exceptions

import docutils.utils
import docutils.parsers.rst.states as states

CURRENT_GL_VERSION = '4.6'
CURRENT_VK_VERSION = '1.4'

TEMPLATE = Template(textwrap.dedent("""\
    ${header}
    ${header_underline}

    %if not bugfix:
    Mesa ${this_version} is a new development release. People who are concerned
    with stability and reliability should stick with a previous release or
    wait for Mesa ${this_version[:-1]}1.
    %else:
    Mesa ${this_version} is a bug fix release which fixes bugs found since the ${previous_version} release.
    %endif

    Mesa ${this_version} implements the OpenGL ${gl_version} API, but the version reported by
    glGetString(GL_VERSION) or glGetIntegerv(GL_MAJOR_VERSION) /
    glGetIntegerv(GL_MINOR_VERSION) depends on the particular driver being used.
    Some drivers don't support all the features required in OpenGL ${gl_version}. OpenGL
    ${gl_version} is **only** available if requested at context creation.
    Compatibility contexts may report a lower version depending on each driver.

    Mesa ${this_version} implements the Vulkan ${vk_version} API, but the version reported by
    the apiVersion property of the VkPhysicalDeviceProperties struct
    depends on the particular driver being used.

    SHA checksums
    -------------

    ::

        TBD.


    New features
    ------------

    %for f in features:
    - ${rst_escape(f)}
    %endfor


    Bug fixes
    ---------

    %for b in bugs:
    - ${rst_escape(b)}
    %endfor


    Changes
    -------
    %for c, author_line in changes:
      %if author_line:

    ${rst_escape(c)}

      %else:
    - ${rst_escape(c)}
      %endif
    %endfor
    """))


# copied from https://docutils.sourceforge.io/sandbox/xml2rst/xml2rstlib/markup.py
class Inliner(states.Inliner):
    """
    Recognizer for inline markup. Derive this from the original inline
    markup parser for best results.
    """

    # Copy static attributes from super class
    vars().update(vars(states.Inliner))

    def quoteInline(self, text):
        """
        `text`: ``str``
          Return `text` with inline markup quoted.
        """
        # Method inspired by `states.Inliner.parse`
        self.document = docutils.utils.new_document("<string>")
        self.document.settings.trim_footnote_reference_space = False
        self.document.settings.character_level_inline_markup = False
        self.document.settings.pep_references = False
        self.document.settings.rfc_references = False

        self.init_customizations(self.document.settings)

        self.reporter = self.document.reporter
        self.reporter.stream = None
        self.language = None
        self.parent = self.document
        remaining = docutils.utils.escape2null(text)
        checked = ""
        processed = []
        unprocessed = []
        messages = []
        while remaining:
            original = remaining
            match = self.patterns.initial.search(remaining)
            if match:
                groups = match.groupdict()
                method = self.dispatch[groups['start'] or groups['backquote']
                                       or groups['refend'] or groups['fnend']]
                before, inlines, remaining, sysmessages = method(self, match, 0)
                checked += before
                if inlines:
                    assert len(inlines) == 1, "More than one inline found"
                    inline = original[len(before)
                                      :len(original) - len(remaining)]
                    rolePfx = re.search("^:" + self.simplename + ":(?=`)",
                                        inline)
                    refSfx = re.search("_+$", inline)
                    if rolePfx:
                        # Prefixed roles need to be quoted in the middle
                        checked += (inline[:rolePfx.end()] + "\\"
                                    + inline[rolePfx.end():])
                    elif refSfx and not re.search("^`", inline):
                        # Pure reference markup needs to be quoted at the end
                        checked += (inline[:refSfx.start()] + "\\"
                                    + inline[refSfx.start():])
                    else:
                        # Quote other inlines by prefixing
                        checked += "\\" + inline
            else:
                checked += remaining
                break
        # Quote all original backslashes
        checked = re.sub('\x00', "\\\x00", checked)
        checked = re.sub('@', '\\@', checked)
        return docutils.utils.unescape(checked, 1)

inliner = Inliner();


async def gather_commits(version: str) -> str:
    p = await asyncio.create_subprocess_exec(
        'git', 'log', '--oneline', f'mesa-{version}..', '-i', '--grep', r'\(Closes\|Fixes\): \(https\|#\).*',
        stdout=asyncio.subprocess.PIPE)
    out, _ = await p.communicate()
    assert p.returncode == 0, f"git log didn't work: {version}"
    return out.decode().strip()


async def parse_issues(commits: str) -> typing.List[str]:
    issues: typing.List[str] = []
    for commit in commits.split('\n'):
        sha, message = commit.split(maxsplit=1)
        p = await asyncio.create_subprocess_exec(
            'git', 'log', '--max-count', '1', r'--format=%b', sha,
            stdout=asyncio.subprocess.PIPE)
        _out, _ = await p.communicate()
        out = _out.decode().split('\n')

        for line in reversed(out):
            if not line.lower().startswith(('closes:', 'fixes:')):
                continue
            bug = line.split(':', 1)[1].strip()
            if (bug.startswith('https://gitlab.freedesktop.org/mesa/mesa')
                # Avoid parsing "merge_requests" URL. Note that a valid issue
                # URL may or may not contain the "/-/" text, so we check if
                # the word "issues" is contained in URL.
                and '/issues' in bug):
                # This means we have a bug in the form "Closes: https://..."
                issues.append(os.path.basename(urllib.parse.urlparse(bug).path))
            elif ',' in bug:
                multiple_bugs = [b.strip().lstrip('#') for b in bug.split(',')]
                if not all(b.isdigit() for b in multiple_bugs):
                    # this is likely a "Fixes" tag that refers to a commit name
                    continue
                issues.extend(multiple_bugs)
            elif bug.startswith('#'):
                issues.append(bug.lstrip('#'))

    return issues


async def gather_bugs(version: str) -> typing.List[str]:
    commits = await gather_commits(version)
    if commits:
        issues = await parse_issues(commits)
    else:
        issues = []

    loop = asyncio.get_event_loop()
    async with aiohttp.ClientSession(loop=loop) as session:
        results = await asyncio.gather(*[get_bug(session, i) for i in issues])
    # Remove duplicates.
    bugs = sorted(set(results))
    if not bugs:
        bugs = ['None']
    return bugs


async def get_bug(session: aiohttp.ClientSession, bug_id: str) -> str:
    """Query gitlab to get the name of the issue that was closed."""
    # Mesa's gitlab id is 176,
    url = 'https://gitlab.freedesktop.org/api/v4/projects/176/issues'
    params = {'iids[]': bug_id}
    async with session.get(url, params=params) as response:
        content = await response.json()
    if not content:
        # issues marked as "confidential" look like "404" page for
        # unauthorized users
        return f'Confidential issue #{bug_id}'
    else:
        return content[0]['title']


async def get_shortlog(version: str) -> str:
    """Call git shortlog."""
    p = await asyncio.create_subprocess_exec('git', 'shortlog', f'mesa-{version}..',
                                             stdout=asyncio.subprocess.PIPE)
    out, _ = await p.communicate()
    assert p.returncode == 0, 'error getting shortlog'
    assert out is not None, 'just for mypy'
    return out.decode()


def walk_shortlog(log: str) -> typing.Generator[typing.Tuple[str, bool], None, None]:
    for l in log.split('\n'):
        if l.startswith(' '): # this means we have a patch description
            yield l.lstrip(), False
        elif l.strip():
            yield l, True


def calculate_next_version(version: str, is_point: bool) -> str:
    """Calculate the version about to be released."""
    if '-' in version:
        version = version.split('-')[0]
    if is_point:
        base = version.split('.')
        base[2] = str(int(base[2]) + 1)
        return '.'.join(base)
    return version


def calculate_previous_version(version: str, is_point: bool) -> str:
    """Calculate the previous version to compare to.

    In the case of -rc to final that version is the previous .0 release,
    (19.3.0 in the case of 20.0.0, for example). for point releases that is
    the last point release. This value will be the same as the input value
    for a point release, but different for a major release.
    """
    if '-' in version:
        version = version.split('-')[0]
    if is_point:
        return version
    base = version.split('.')
    if base[1] == '0':
        base[0] = str(int(base[0]) - 1)
        base[1] = '3'
    else:
        base[1] = str(int(base[1]) - 1)
    return '.'.join(base)


def get_features(is_point_release: bool) -> typing.Generator[str, None, None]:
    p = pathlib.Path('docs') / 'relnotes' / 'new_features.txt'
    if p.exists() and p.stat().st_size > 0:
        if is_point_release:
            print("WARNING: new features being introduced in a point release", file=sys.stderr)
        with p.open('rt') as f:
            for line in f:
                yield line.rstrip()
        p.unlink()
        subprocess.run(['git', 'add', p])
    else:
        yield "None"


def update_release_notes_index(version: str) -> None:
    relnotes_index_path = pathlib.Path('docs') / 'relnotes.rst'

    with relnotes_index_path.open('r') as f:
        relnotes = f.readlines()

    new_relnotes = []
    first_list = True
    second_list = True
    for line in relnotes:
        if first_list and line.startswith('-'):
            first_list = False
            new_relnotes.append(f'-  :doc:`{version} release notes <relnotes/{version}>`\n')
        if (not first_list and second_list and
            re.match(r'   \d+.\d+(.\d+)? <relnotes/\d+.\d+(.\d+)?>', line)):
            second_list = False
            new_relnotes.append(f'   {version} <relnotes/{version}>\n')
        new_relnotes.append(line)

    with relnotes_index_path.open('w', encoding='utf-8') as f:
        for line in new_relnotes:
            f.write(line)

    subprocess.run(['git', 'add', relnotes_index_path])


async def main() -> None:
    v = pathlib.Path('VERSION')
    with v.open('rt') as f:
        raw_version = f.read().strip()
    is_point_release = '-rc' not in raw_version
    assert '-devel' not in raw_version, 'Do not run this script on -devel'
    version = raw_version.split('-')[0]
    previous_version = calculate_previous_version(version, is_point_release)
    this_version = calculate_next_version(version, is_point_release)
    today = datetime.date.today()
    header = f'Mesa {this_version} Release Notes / {today}'
    header_underline = '=' * len(header)

    shortlog, bugs = await asyncio.gather(
        get_shortlog(previous_version),
        gather_bugs(previous_version),
    )

    final = pathlib.Path('docs') / 'relnotes' / f'{this_version}.rst'
    with final.open('wt', encoding='utf-8') as f:
        try:
            f.write(TEMPLATE.render(
                bugfix=is_point_release,
                bugs=bugs,
                changes=walk_shortlog(shortlog),
                features=get_features(is_point_release),
                gl_version=CURRENT_GL_VERSION,
                this_version=this_version,
                header=header,
                header_underline=header_underline,
                previous_version=previous_version,
                vk_version=CURRENT_VK_VERSION,
                rst_escape=inliner.quoteInline,
            ))
        except:
            print(exceptions.text_error_template().render())
            return

    subprocess.run(['git', 'add', final])

    update_release_notes_index(this_version)

    subprocess.run(['git', 'commit', '-m',
                    f'docs: add release notes for {this_version}'])


if __name__ == "__main__":
    loop = asyncio.new_event_loop()
    loop.run_until_complete(main())




# (C) Copyright IBM Corporation 2004, 2005
# All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# on the rights to use, copy, modify, merge, publish, distribute, sub
# license, and/or sell copies of the Software, and to permit persons to whom
# the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT.  IN NO EVENT SHALL
# IBM AND/OR ITS SUPPLIERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
# Authors:
#    Ian Romanick <idr@us.ibm.com>

import gl_XML
import license
import sys, getopt, string


class glx_item_factory(gl_XML.gl_item_factory):
    """Factory to create GLX protocol oriented objects derived from gl_item."""

    def create_function(self, element, context):
        return glx_function(element, context)

    def create_enum(self, element, context, category):
        return glx_enum(element, context, category)

    def create_api(self, pointer_type):
        return glx_api(self, pointer_type)


class glx_enum(gl_XML.gl_enum):
    def __init__(self, element, context, category):
        gl_XML.gl_enum.__init__(self, element, context, category)

        self.functions = {}

        for child in element:
            if child.tag == "size":
                n = child.get( "name" )
                c = child.get( "count" )
                m = child.get( "mode", "set" )

                if not c:
                    c = self.default_count
                else:
                    c = int(c)

                if m == "get":
                    mode = 0
                else:
                    mode = 1

                if n not in self.functions:
                    self.functions[ n ] = [c, mode]

        return


class glx_function(gl_XML.gl_function):
    def __init__(self, element, context):
        self.glx_rop = 0
        self.glx_sop = 0
        self.glx_vendorpriv = 0

        self.glx_vendorpriv_names = []

        # If this is set to true, it means that GLdouble parameters should be
        # written to the GLX protocol packet in the order they appear in the
        # prototype.  This is different from the "classic" ordering.  In the
        # classic ordering GLdoubles are written to the protocol packet first,
        # followed by non-doubles.  NV_vertex_program was the first extension
        # to break with this tradition.

        self.glx_doubles_in_order = 0

        self.vectorequiv = None
        self.output = None
        self.can_be_large = 0
        self.reply_always_array = 0
        self.dimensions_in_reply = 0
        self.img_reset = None

        self.client_handcode = 0
        self.ignore = 0

        self.count_parameter_list = []
        self.counter_list = []
        self.parameters_by_name = {}
        self.offsets_calculated = 0

        gl_XML.gl_function.__init__(self, element, context)
        return


    def process_element(self, element):
        gl_XML.gl_function.process_element(self, element)

        # If the function already has a vector equivalent set, don't
        # set it again.  This can happen if an alias to a function
        # appears after the function that it aliases.

        if not self.vectorequiv:
            self.vectorequiv = element.get("vectorequiv")


        name = element.get("name")
        if name == self.name:
            for param in self.parameters:
                self.parameters_by_name[ param.name ] = param

                if len(param.count_parameter_list):
                    self.count_parameter_list.extend( param.count_parameter_list )

                if param.counter and param.counter not in self.counter_list:
                    self.counter_list.append(param.counter)


        for child in element:
            if child.tag == "glx":
                rop = child.get( 'rop' )
                sop = child.get( 'sop' )
                vop = child.get( 'vendorpriv' )

                if rop:
                    self.glx_rop = int(rop)

                if sop:
                    self.glx_sop = int(sop)

                if vop:
                    self.glx_vendorpriv = int(vop)
                    self.glx_vendorpriv_names.append(name)

                self.img_reset = child.get( 'img_reset' )

                # The 'handcode' attribute can be one of 'true',
                # 'false', 'client', or 'server'.

                handcode = child.get( 'handcode', 'false' )
                if handcode == "false":
                    self.client_handcode = 0
                elif handcode == "true":
                    self.client_handcode = 1
                elif handcode == "client":
                    self.client_handcode = 1
                elif handcode == "server":
                    self.client_handcode = 0
                else:
                    raise RuntimeError('Invalid handcode mode "%s" in function "%s".' % (handcode, self.name))

                self.ignore               = gl_XML.is_attr_true( child, 'ignore' )
                self.can_be_large         = gl_XML.is_attr_true( child, 'large' )
                self.glx_doubles_in_order = gl_XML.is_attr_true( child, 'doubles_in_order' )
                self.reply_always_array   = gl_XML.is_attr_true( child, 'always_array' )
                self.dimensions_in_reply  = gl_XML.is_attr_true( child, 'dimensions_in_reply' )


        # Do some validation of the GLX protocol information.  As
        # new tests are discovered, they should be added here.

        for param in self.parameters:
            if param.is_output and self.glx_rop != 0:
                raise RuntimeError("Render / RenderLarge commands cannot have outputs (%s)." % (self.name))

        return


    def variable_length_parameter(self):
        for param in self.parameters:
            if not param.is_output:
                if param.counter or len(param.count_parameter_list):
                    return param

        return None


    def calculate_offsets(self):
        if not self.offsets_calculated:
            # Calculate the offset of the first function parameter
            # in the GLX command packet.  This byte offset is
            # measured from the end of the Render / RenderLarge
            # header.  The offset for all non-pixel commends is
            # zero.  The offset for pixel commands depends on the
            # number of dimensions of the pixel data.

            if len(self.images) and not self.images[0].is_output:
                [dim, _, _, _, _] = self.images[0].get_dimensions()

                # The base size is the size of the pixel pack info
                # header used by images with the specified number
                # of dimensions.

                if dim <=  2:
                    offset = 20
                elif dim <= 4:
                    offset = 36
                else:
                    raise RuntimeError('Invalid number of dimensions %u for parameter "%s" in function "%s".' % (dim, self.image.name, self.name))
            else:
                offset = 0

            for param in self.parameterIterateGlxSend():
                if param.img_null_flag:
                    offset += 4

                if param.name != self.img_reset:
                    param.offset = offset
                    if not param.is_variable_length() and not param.is_client_only:
                        offset += param.size()

                if self.pad_after( param ):
                    offset += 4


            self.offsets_calculated = 1
        return


    def parameterIterateGlxSend(self, include_variable_parameters = 1):
        """Create an iterator for parameters in GLX request order."""

        # The parameter lists are usually quite short, so it's easier
        # (i.e., less code) to just generate a new list with the
        # required elements than it is to create a new iterator class.

        temp = [ [],  [], [] ]
        for param in self.parameters:
            if param.is_output: continue

            if param.is_variable_length():
                temp[2].append( param )
            elif not self.glx_doubles_in_order and param.is_64_bit():
                temp[0].append( param )
            else:
                temp[1].append( param )

        parameters = temp[0]
        parameters.extend( temp[1] )
        if include_variable_parameters:
            parameters.extend( temp[2] )
        return iter(parameters)


    def parameterIterateCounters(self):
        temp = []
        for name in self.counter_list:
            temp.append( self.parameters_by_name[ name ] )

        return iter(temp)


    def parameterIterateOutputs(self):
        temp = []
        for p in self.parameters:
            if p.is_output:
                temp.append( p )

        return temp


    def command_fixed_length(self):
        """Return the length, in bytes as an integer, of the
        fixed-size portion of the command."""

        if len(self.parameters) == 0:
            return 0

        self.calculate_offsets()

        size = 0
        for param in self.parameterIterateGlxSend(0):
            if param.name != self.img_reset and not param.is_client_only:
                if size == 0:
                    size = param.offset + param.size()
                else:
                    size += param.size()

                if self.pad_after( param ):
                    size += 4

        for param in self.images:
            if param.img_null_flag or param.is_output:
                size += 4

        return size


    def command_variable_length(self):
        """Return the length, as a string, of the variable-sized
        portion of the command."""

        size_string = ""
        for p in self.parameterIterateGlxSend():
            if (not p.is_output) and (p.is_variable_length() or p.is_image()):
                # FIXME Replace the 1 in the size_string call
                # FIXME w/0 to eliminate some un-needed parnes
                # FIXME This would already be done, but it
                # FIXME adds some extra diffs to the generated
                # FIXME code.

                size_string = size_string + " + safe_pad(%s)" % (p.size_string(1))

        return size_string


    def command_length(self):
        size = self.command_fixed_length()

        if self.glx_rop != 0:
            size += 4

        size = ((size + 3) & ~3)
        return "%u%s" % (size, self.command_variable_length())


    def opcode_value(self):
        """Get the unique protocol opcode for the glXFunction"""

        if (self.glx_rop == 0) and self.vectorequiv:
            equiv = self.context.functions_by_name[ self.vectorequiv ]
            self.glx_rop = equiv.glx_rop


        if self.glx_rop != 0:
            return self.glx_rop
        elif self.glx_sop != 0:
            return self.glx_sop
        elif self.glx_vendorpriv != 0:
            return self.glx_vendorpriv
        else:
            return -1


    def opcode_rop_basename(self):
        """Return either the name to be used for GLX protocol enum.

        Returns either the name of the function or the name of the
        name of the equivalent vector (e.g., glVertex3fv for
        glVertex3f) function."""

        if self.vectorequiv == None:
            return self.name
        else:
            return self.vectorequiv


    def opcode_name(self):
        """Get the unique protocol enum name for the glXFunction"""

        if (self.glx_rop == 0) and self.vectorequiv:
            equiv = self.context.functions_by_name[ self.vectorequiv ]
            self.glx_rop = equiv.glx_rop
            self.glx_doubles_in_order = equiv.glx_doubles_in_order


        if self.glx_rop != 0:
            return "X_GLrop_%s" % (self.opcode_rop_basename())
        elif self.glx_sop != 0:
            return "X_GLsop_%s" % (self.name)
        elif self.glx_vendorpriv != 0:
            return "X_GLvop_%s" % (self.name)
        else:
            raise RuntimeError('Function "%s" has no opcode.' % (self.name))


    def opcode_vendor_name(self, name):
        if name in self.glx_vendorpriv_names:
            return "X_GLvop_%s" % (name)
        else:
            raise RuntimeError('Function "%s" has no VendorPrivate opcode.' % (name))


    def opcode_real_name(self):
        """Get the true protocol enum name for the GLX opcode

        Behaves similarly to opcode_name, except for
        X_GLXVendorPrivate and X_GLXVendorPrivateWithReply commands.
        In these cases the string 'X_GLXVendorPrivate' or
        'X_GLXVendorPrivateWithReply' is returned.  For other
        single or render commands 'X_GLsop' or 'X_GLrop' plus the
        name of the function returned."""

        if self.glx_vendorpriv != 0:
            if self.needs_reply():
                return "X_GLXVendorPrivateWithReply"
            else:
                return "X_GLXVendorPrivate"
        else:
            return self.opcode_name()


    def needs_reply(self):
        try:
            x = self._needs_reply
        except Exception:
            x = 0
            if self.return_type != 'void':
                x = 1

            for param in self.parameters:
                if param.is_output:
                    x = 1
                    break

            self._needs_reply = x

        return x


    def pad_after(self, p):
        """Returns the name of the field inserted after the
        specified field to pad out the command header."""

        for image in self.images:
            if image.img_pad_dimensions:
                if not image.height:
                    if p.name == image.width:
                        return "height"
                    elif p.name == image.img_xoff:
                        return "yoffset"
                elif not image.extent:
                    if p.name == image.depth:
                        # Should this be "size4d"?
                        return "extent"
                    elif p.name == image.img_zoff:
                        return "woffset"

        return None


    def has_different_protocol(self, name):
        """Returns true if the named version of the function uses different protocol from the other versions.

        Some functions, such as glDeleteTextures and
        glDeleteTexturesEXT are functionally identical, but have
        different protocol.  This function returns true if the
        named function is an alias name and that named version uses
        different protocol from the function that is aliased.
        """

        return (name in self.glx_vendorpriv_names) and self.glx_sop


    def static_glx_name(self, name, gl_symbols):
        if self.has_different_protocol(name):
            for n in self.glx_vendorpriv_names:
                if n in gl_symbols:
                    return n

        return self.static_name(name, gl_symbols)


    def client_supported_for_indirect(self):
        """Returns true if the function is supported on the client
        side for indirect rendering."""

        return (not self.ignore and self.exec_flavor != 'skip' and (self.offset != -1) and
                (self.glx_rop or self.glx_sop or self.glx_vendorpriv or self.vectorequiv or
                 self.client_handcode))


class glx_function_iterator(object):
    """Class to iterate over a list of glXFunctions"""

    def __init__(self, context):
        self.iterator = context.functionIterateByOffset()
        return


    def __iter__(self):
        return self


    def __next__(self):
        while True:
            f = next(self.iterator)

            if f.client_supported_for_indirect():
                return f

    next = __next__


class glx_api(gl_XML.gl_api):
    def functionIterateGlx(self):
        return glx_function_iterator(self)




COPYRIGHT=u"""
/* Copyright 2024 Valve Corporation
 * Copyright 2021 Intel Corporation
 * SPDX-License-Identifier: MIT
 */
"""

import argparse
from vk_physical_device_features_gen import get_renamed_feature
import os
import sys
import xml.etree.ElementTree as et

import mako
from mako.template import Template

TEMPLATE_C = Template(COPYRIGHT + """
/* This file generated from ${filename}, don't edit directly. */

#include "vk_physical_device.h"
#include "vk_instance.h"
#include "vk_shader.h"

/* for spirv_supported_capabilities */
#include "compiler/spirv/spirv_info.h"

struct spirv_capabilities
vk_physical_device_get_spirv_capabilities(const struct vk_physical_device *pdev)
{
   const struct vk_features *f = &pdev->supported_features;
   const struct vk_device_extension_table *e = &pdev->supported_extensions;
   const struct vk_properties *p = &pdev->properties;
   uint32_t api_version = pdev->instance->app_info.api_version;

   struct spirv_capabilities caps = { false, };

   /* We |= for everything because some caps have multiple names but the
    * same enum value and they sometimes have different enables in the
    * Vulkan spec.  To handle this, we just | all the enables together.
    */
% for cap in caps:
    caps.${cap} |= ${' | '.join(caps[cap])};
% endfor

   return caps;
}
""")

# These don't exist in the SPIR-V headers for one reason or another.
NON_EXISTANT_CAPS = [
    # This isn't a cap, it's an execution mode.
    #
    # https://gitlab.khronos.org/vulkan/vulkan/-/merge_requests/6618
    'MaximallyReconvergesKHR',

    # This extension got published but never got merged to SPIRV-Headers
    #
    # https://gitlab.khronos.org/spirv/spirv-extensions/-/merge_requests/238
    'ClusterCullingShadingHUAWEI',

    # Exclude the one beta cap.
    'ShaderEnqueueAMDX',
]

def process_enable(enab):
    attrib = enab.attrib

    if 'property' in attrib:
        if attrib['value'] == 'VK_TRUE':
            return f"p->{attrib['member']}"
        else:
            return f"(p->{attrib['member']} & {attrib['value']})"
    elif 'extension' in attrib:
        return f"e->{attrib['extension'].removeprefix('VK_')}"
    elif 'feature' in attrib:
        feat = get_renamed_feature(attrib['struct'], attrib['feature'])
        return f"f->{feat}"
    else:
        version = attrib['version']
        return f"(api_version >= VK_API_{version.removeprefix('VK_')})"

def get_capabilities(doc, beta):
    caps = {}

    for cap in doc.findall('./spirvcapabilities/spirvcapability'):
        name = cap.attrib['name']
        if name in NON_EXISTANT_CAPS:
            continue

        enables = cap.findall('enable')
        lst = caps.setdefault(name, [])
        lst += [process_enable(x) for x in enables]

    # Remove duplicates
    for cap in caps:
        caps[cap] = list(dict.fromkeys(caps[cap]))

    return caps


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--out-c', required=True, help='Output C file.')
    parser.add_argument('--beta', required=True, help='Enable beta extensions.')
    parser.add_argument('--xml', required=True, help='Vulkan API XML file.')
    args = parser.parse_args()

    environment = {
        'filename': os.path.basename(__file__),
        'caps': get_capabilities(et.parse(args.xml), args.beta),
    }

    try:
        with open(args.out_c, 'w', encoding='utf-8') as f:
            f.write(TEMPLATE_C.render(**environment))
    except Exception:
        # In the event there's an error, this uses some helpers from mako
        # to print a useful stack trace and prints it, then exits with
        # status 1, if python is run with debug; otherwise it just raises
        # the exception
        print(mako.exceptions.text_error_template().render(), file=sys.stderr)
        sys.exit(1)

if __name__ == '__main__':
    main()



#!/usr/bin/python3
#
# Copyright © 2024-2025 Tomeu Vizoso
#
# SPDX-License-Identifier: MIT

import sys
import os
import argparse
import struct
from gen_parser import Parser, Reg, Enum, mask, Error


def main():
	parser = argparse.ArgumentParser()
	parser.add_argument('--xml', type=str, required=True)
	parser.add_argument('--dump', type=str, required=True)

	args = parser.parse_args()

	p = Parser()

	try:
		p.parse("", args.xml)
	except Error as e:
		print(e, file=sys.stderr)
		exit(1)

	regs = {}
	for e in p.file:
		if isinstance(e, Reg):
			regs[e.offset] = e

	domains = {}
	for e in p.file:
		if isinstance(e, Enum):
			if e.name == "target":
				for name, val in e.values:
					domains[name] = val

	f = open(args.dump, mode='rb')
	for i in range(0, os.path.getsize(args.dump) // 8):
		cmd = f.read(8)
		(offset, value, target) = struct.unpack("<hIh", cmd)
		if offset in regs.keys():
			reg = regs[offset]

			if (target & 0xfffffffe) != domains[reg.domain]:
				print("WARNING: target 0x%x doesn't match register's domain 0x%x" % (target, domains[reg.domain]))

			print("EMIT(REG_%s, " % regs[offset].full_name.upper(), end="")
			first = True
			if value == 0 or len(reg.bitset.fields) == 1:
				print("0x%x" % value, end="")
			else:
				for field in reg.bitset.fields:
					if field.type == "boolean":
						if 1 << field.high & value:
							if not first:
								print(" | ", end="")
							print("%s_%s" % (reg.full_name.upper(), field.name.upper()), end="")
							first = False
					elif field.type == "uint":
						field_value = (value & mask(field.low, field.high)) >> field.low
						if field_value != 0:
							if not first:
								print(" | ", end="")
							print("%s_%s(%d)" % (reg.full_name.upper(), field.name.upper(), field_value), end="")
							first = False
			print(");")
		else:
			print("%x %x %x" % (target, offset, value))

if __name__ == '__main__':
	main()



#! /usr/bin/env python3
#
# Copyright © 2024 Collabora Ltd. and Red Hat Inc.
# SPDX-License-Identifier: MIT

import argparse
import os.path
import re
import sys

from collections import namedtuple
from mako.template import Template

import util


TEMPLATE_RS = Template("""\
// Copyright © 2024 Collabora Ltd. and Red Hat Inc.
// SPDX-License-Identifier: MIT

// This file is generated by struct_parser.py. DO NOT EDIT!

#![allow(non_snake_case)]

use std::ops::Range;

% for s in structs:
pub const ${s.name}_MAX_BIT: usize = ${s.max_bit};
    % for f in s.fields:
        % if f.stride:
#[inline]
pub const fn ${s.name}_${f.name}(i: usize) -> Range<usize> {
        % if f.stride == 1:
    (i + ${f.lo})..(i + ${f.hi + 1})
        % else:
    (i * ${f.stride} + ${f.lo})..(i * ${f.stride} + ${f.hi + 1})
        % endif
}
        % else:
pub const ${s.name}_${f.name}: Range<usize> = ${f.lo}..${f.hi + 1};
        % endif:
        % for e in f.enums:
pub const ${s.name}_${f.name}_${e.name}: u32 = ${e.value};
        % endfor
    % endfor
% endfor
""")

STRUCTS = [
    'SPHV3',
    'SPHV4',
    'TEXHEADV2',
    'TEXHEADV3',
    'TEXHEAD_BL',
    'TEXHEAD_1D',
    'TEXHEAD_PITCH',
    # This one goes last because it's a substring of the others
    'TEXHEAD',
    'TEXSAMP',
    'QMDV00_06',
    'QMDV01_06',
    'QMDV01_07',
    'QMDV02_01',
    'QMDV02_02',
    'QMDV02_03',
    'QMDV02_04',
    'QMDV03_00',
    'QMDV04_00',
    'QMDV05_00',
]

Enum = namedtuple('Enum', ['name', 'value'])

class Field(object):
    def __init__(self, name, lo, hi, stride=0):
        self.name = name
        self.lo = lo
        self.hi = hi
        self.stride = stride
        self.enums = []

    def add_enum(self, name, value):
        self.enums.append(Enum(name, value))

class Struct(object):
    def __init__(self, name):
        self.name = name
        self.fields = []
        self.max_bit = 0

    def add_field(self, name, lo, hi, stride=0):
        self.fields.append(Field(name, lo, hi, stride))
        self.max_bit = max(self.max_bit, hi)

DRF_RE = re.compile(r'(?P<hi>[0-9]+):(?P<lo>[0-9]+)')
FIELD_NAME_RE = re.compile(r'_?(?P<dw>[0-9]+)?_?(?P<name>.*)')
MW_RE = re.compile(r'MW\((?P<hi>[0-9]+):(?P<lo>[0-9]+)\)')
MW_ARR_RE = re.compile(r'MW\(\((?P<hi>\d+)\+\(i\)\*(?P<stride>\d+)\):\((?P<lo>[0-9]+)\+\(i\)\*(?P=stride)\)\)')

def parse_header(nvcl, file):
    structs = {}
    for line in file:
        line = line.strip().split()
        if not line:
            continue

        if line[0] != '#define':
            continue

        if not line[1].startswith(nvcl):
            continue

        name = line[1][(len(nvcl)+1):]

        struct = None
        for s in STRUCTS:
            if name.startswith(s):
                if s not in structs:
                    structs[s] = Struct(s)
                struct = structs[s]
                name = name[len(s):]
                break

        if struct is None:
            continue

        name_m = FIELD_NAME_RE.match(name)
        name = name_m.group('name')

        drf = DRF_RE.match(line[2])
        mw = MW_RE.match(line[2])
        mw_arr = MW_ARR_RE.match(line[2])
        if drf:
            dw = int(name_m.group('dw'))
            lo = int(drf.group('lo')) + dw * 32
            hi = int(drf.group('hi')) + dw * 32
            struct.add_field(name, lo, hi)
        elif mw:
            lo = int(mw.group('lo'))
            hi = int(mw.group('hi'))
            struct.add_field(name, lo, hi)
        elif mw_arr:
            lo = int(mw_arr.group('lo'))
            hi = int(mw_arr.group('hi'))
            stride = int(mw_arr.group('stride'))
            assert name.endswith('(i)')
            struct.add_field(name.removesuffix('(i)'), lo, hi, stride)
        else:
            for f in struct.fields:
                if name.startswith(f.name + '_'):
                    name = name[(len(f.name)+1):]
                    f.add_enum(name, line[2])

    return list(structs.values())

NVCL_RE = re.compile(r'cl(?P<clsver>[0-9a-f]{4}).*')

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('--out-rs', required=True, help='Output Rust file.')
    parser.add_argument('--in-h',
                        help='Input class header file.',
                        required=True)
    args = parser.parse_args()

    clheader = os.path.basename(args.in_h)
    nvcl = NVCL_RE.match(clheader).group('clsver')
    nvcl = nvcl.upper()
    nvcl = "NV" + nvcl

    with open(args.in_h, 'r', encoding='utf-8') as f:
        structs = parse_header(nvcl, f)

    util.write_template_rs(args.out_rs, TEMPLATE_RS, dict(structs=structs))


if __name__ == '__main__':
    main()



# Copyright (C) 2021 Collabora, Ltd.
# Copyright (C) 2016 Intel Corporation
# SPDX-License-Identifier: MIT

import argparse
import sys
import math

a = 'a'
b = 'b'
c = 'c'
d = 'd'

# In general, bcsel is cheaper than bitwise arithmetic on Mali. On
# Bifrost, we can implement bcsel as either CSEL or MUX to schedule to either
# execution unit. On Valhall, bitwise arithmetic may be on the SFU whereas MUX
# is on the higher throughput CVT unit. We get a zero argument for free relative
# to the bitwise op, which would be LSHIFT_* internally taking a zero anyway.
#
# As such, it's beneficial to reexpress bitwise arithmetic of booleans as bcsel.
opt_bool_bitwise = [
    (('iand', 'a@1', 'b@1'), ('bcsel', a, b, False)),
    (('ior', 'a@1', 'b@1'), ('bcsel', a, a, b)),
    (('iand', 'a@1', ('inot', 'b@1')), ('bcsel', b, 0, a)),
    (('ior', 'a@1', ('inot', 'b@1')), ('bcsel', b, a, True)),
]

algebraic_late = [
    # Canonical form. The scheduler will convert back if it makes sense.
    (('fmul', a, 2.0), ('fadd', a, a)),

    # Fuse Mali-specific clamps
    (('fmin', ('fmax', a, -1.0), 1.0), ('fsat_signed', a)),
    (('fmax', ('fmin', a, 1.0), -1.0), ('fsat_signed', a)),
    (('fmax', a, 0.0), ('fclamp_pos', a)),

    (('b32csel', 'b@32', ('iadd', 'a@32', 1), a), ('iadd', a, ('b2i32', b))),

    # We don't have an 8-bit CSEL, so this is the best we can do.
    # Note that we use 8-bit booleans internally to preserve vectorization.
    (('imin', 'a@8', 'b@8'), ('b8csel', ('ilt8', a, b), a, b)),
    (('imax', 'a@8', 'b@8'), ('b8csel', ('ilt8', a, b), b, a)),
    (('umin', 'a@8', 'b@8'), ('b8csel', ('ult8', a, b), a, b)),
    (('umax', 'a@8', 'b@8'), ('b8csel', ('ult8', a, b), b, a)),

    # Floats are at minimum 16-bit, which means when converting to an 8-bit
    # integer, the vectorization changes. So there's no one-shot hardware
    # instruction for f2i8. Instead, lower to two NIR instructions that map
    # directly to the hardware.
    (('f2i8', a), ('i2i8', ('f2i16', a))),
    (('f2u8', a), ('u2u8', ('f2u16', a))),

    # XXX: Duplicate of nir_lower_pack
    (('unpack_64_2x32', a), ('vec2', ('unpack_64_2x32_split_x', a),
                                     ('unpack_64_2x32_split_y', a))),

    # On v11+, all non integer variant to convert to F32 are gone except for S32_TO_F32.
    (('i2f32', 'a@8'), ('i2f32', ('i2i32', a)), 'gpu_arch >= 11'),
    (('i2f32', 'a@16'), ('i2f32', ('i2i32', a)), 'gpu_arch >= 11'),
    (('u2f32', 'a@8'), ('u2f32', ('u2u32', a)), 'gpu_arch >= 11'),
    (('u2f32', 'a@16'), ('u2f32', ('u2u32', a)), 'gpu_arch >= 11'),

    # On v11+, all non integer variant to convert to F16 are gone except for S32_TO_F32.
    (('i2f16', 'a'), ('f2f16', ('i2f32', ('i2i32', a))), 'gpu_arch >= 11'),
    (('u2f16', 'a'), ('f2f16', ('u2f32', ('u2u32', a))), 'gpu_arch >= 11'),

    # We don't have S32_TO_F16 on any arch
    (('i2f16', 'a@32'), ('f2f16', ('i2f32', a))),
    (('u2f16', 'a@32'), ('f2f16', ('u2f32', a))),

    # On v11+, V2F16_TO_V2S16 / V2F16_TO_V2U16 are gone
    (('f2i16', 'a@16'), ('f2i16', ('f2f32', a)), 'gpu_arch >= 11'),
    (('f2u16', 'a@16'), ('f2u16', ('f2f32', a)), 'gpu_arch >= 11'),

    # On v11+, V2F32_TO_V2S16 is gone
    (('pack_half_2x16_split', a, b), ('pack_32_2x16_split', ('f2f16', a), ('f2f16', b)), 'gpu_arch >= 11'),

    # On v11+, F16_TO_S32/F16_TO_U32 is gone but we still have F32_TO_S32/F32_TO_U32
    (('f2i32', 'a@16'), ('f2i32', ('f2f32', a)), 'gpu_arch >= 11'),
    (('f2u32', 'a@16'), ('f2u32', ('f2f32', a)), 'gpu_arch >= 11'),

    # On v11+, because FROUND.v2f16 is gone we end up with precision issues.
    # We lower ffract here instead to ensure lower_bit_size has been performed.
    (('ffract', a), ('fadd', a, ('fneg', ('ffloor', a))), 'gpu_arch >= 11'),
]

# nir_lower_bool_to_bitsize can generate needless conversions.
for bits in [8, 16, 32]:
    algebraic_late += [
        ((f'i2i{bits}', f'a@{bits}'), a)
    ]

# On v11+, ICMP_OR.v4u8 was removed
for cond in ['ilt', 'ige', 'ieq', 'ine', 'ult', 'uge']:
    convert_8bit = 'u2u8'
    convert_16bit = 'u2u16'

    if cond[0] == 'i':
        convert_8bit = 'i2i8'
        convert_16bit = 'i2i16'

    algebraic_late += [
        ((f'{cond}8', a, b), (convert_8bit, (f'{cond}16', (convert_16bit, a), (convert_16bit, b))), 'gpu_arch >= 11'),
    ]

# Handling all combinations of boolean and float sizes for b2f is nontrivial.
# bcsel has the same problem in more generality; lower b2f to bcsel in NIR to
# reuse the efficient implementations of bcsel. This includes special handling
# to allow vectorization in places the hardware does not directly.
#
# Because this lowering must happen late, NIR won't squash inot in
# automatically. Do so explicitly. (The more specific pattern must be first.)
for bsz in [8, 16, 32]:
    for fsz in [16, 32]:
        if bsz == fsz:
            a_fsz = 'a'
        else:
            a_fsz = (f'i2i{fsz}', a)

        algebraic_late += [
            ((f'b2f{fsz}', ('inot', f'a@{bsz}')), (f'b{fsz}csel', a_fsz, 0.0, 1.0)),
            ((f'b2f{fsz}', f'a@{bsz}'), (f'b{fsz}csel', a_fsz, 1.0, 0.0)),
        ]


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--import-path', required=True)
    args = parser.parse_args()
    sys.path.insert(0, args.import_path)
    run()


def run():
    import nir_algebraic  # pylint: disable=import-error

    print('#include "bifrost_nir.h"')

    print(nir_algebraic.AlgebraicPass("bifrost_nir_opt_boolean_bitwise",
                                      opt_bool_bitwise).render())
    print(nir_algebraic.AlgebraicPass("bifrost_nir_lower_algebraic_late",
                                      algebraic_late,
                                      [
                                          ("unsigned ", "gpu_arch")
                                      ]).render())

if __name__ == '__main__':
    main()




# Copyright (C) 2015 Intel Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice (including the next
# paragraph) shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import os
import sys

# The list functions present in the dispatch table.
all_functions = [
    "NewList",
    "EndList",
    "CallList",
    "CallLists",
    "DeleteLists",
    "GenLists",
    "ListBase",
    "Begin",
    "Bitmap",
    "Color3b",
    "Color3bv",
    "Color3d",
    "Color3dv",
    "Color3f",
    "Color3fv",
    "Color3i",
    "Color3iv",
    "Color3s",
    "Color3sv",
    "Color3ub",
    "Color3ubv",
    "Color3ui",
    "Color3uiv",
    "Color3us",
    "Color3usv",
    "Color4b",
    "Color4bv",
    "Color4d",
    "Color4dv",
    "Color4f",
    "Color4fv",
    "Color4i",
    "Color4iv",
    "Color4s",
    "Color4sv",
    "Color4ub",
    "Color4ubv",
    "Color4ui",
    "Color4uiv",
    "Color4us",
    "Color4usv",
    "EdgeFlag",
    "EdgeFlagv",
    "End",
    "Indexd",
    "Indexdv",
    "Indexf",
    "Indexfv",
    "Indexi",
    "Indexiv",
    "Indexs",
    "Indexsv",
    "Normal3b",
    "Normal3bv",
    "Normal3d",
    "Normal3dv",
    "Normal3f",
    "Normal3fv",
    "Normal3i",
    "Normal3iv",
    "Normal3s",
    "Normal3sv",
    "RasterPos2d",
    "RasterPos2dv",
    "RasterPos2f",
    "RasterPos2fv",
    "RasterPos2i",
    "RasterPos2iv",
    "RasterPos2s",
    "RasterPos2sv",
    "RasterPos3d",
    "RasterPos3dv",
    "RasterPos3f",
    "RasterPos3fv",
    "RasterPos3i",
    "RasterPos3iv",
    "RasterPos3s",
    "RasterPos3sv",
    "RasterPos4d",
    "RasterPos4dv",
    "RasterPos4f",
    "RasterPos4fv",
    "RasterPos4i",
    "RasterPos4iv",
    "RasterPos4s",
    "RasterPos4sv",
    "Rectd",
    "Rectdv",
    "Rectf",
    "Rectfv",
    "Recti",
    "Rectiv",
    "Rects",
    "Rectsv",
    "TexCoord1d",
    "TexCoord1dv",
    "TexCoord1f",
    "TexCoord1fv",
    "TexCoord1i",
    "TexCoord1iv",
    "TexCoord1s",
    "TexCoord1sv",
    "TexCoord2d",
    "TexCoord2dv",
    "TexCoord2f",
    "TexCoord2fv",
    "TexCoord2i",
    "TexCoord2iv",
    "TexCoord2s",
    "TexCoord2sv",
    "TexCoord3d",
    "TexCoord3dv",
    "TexCoord3f",
    "TexCoord3fv",
    "TexCoord3i",
    "TexCoord3iv",
    "TexCoord3s",
    "TexCoord3sv",
    "TexCoord4d",
    "TexCoord4dv",
    "TexCoord4f",
    "TexCoord4fv",
    "TexCoord4i",
    "TexCoord4iv",
    "TexCoord4s",
    "TexCoord4sv",
    "Vertex2d",
    "Vertex2dv",
    "Vertex2f",
    "Vertex2fv",
    "Vertex2i",
    "Vertex2iv",
    "Vertex2s",
    "Vertex2sv",
    "Vertex3d",
    "Vertex3dv",
    "Vertex3f",
    "Vertex3fv",
    "Vertex3i",
    "Vertex3iv",
    "Vertex3s",
    "Vertex3sv",
    "Vertex4d",
    "Vertex4dv",
    "Vertex4f",
    "Vertex4fv",
    "Vertex4i",
    "Vertex4iv",
    "Vertex4s",
    "Vertex4sv",
    "ClipPlane",
    "ColorMaterial",
    "CullFace",
    "Fogf",
    "Fogfv",
    "Fogi",
    "Fogiv",
    "FrontFace",
    "Hint",
    "Lightf",
    "Lightfv",
    "Lighti",
    "Lightiv",
    "LightModelf",
    "LightModelfv",
    "LightModeli",
    "LightModeliv",
    "LineStipple",
    "LineWidth",
    "Materialf",
    "Materialfv",
    "Materiali",
    "Materialiv",
    "PointSize",
    "PolygonMode",
    "PolygonStipple",
    "Scissor",
    "ShadeModel",
    "TexParameterf",
    "TexParameterfv",
    "TexParameteri",
    "TexParameteriv",
    "TexImage1D",
    "TexImage2D",
    "TexEnvf",
    "TexEnvfv",
    "TexEnvi",
    "TexEnviv",
    "TexGend",
    "TexGendv",
    "TexGenf",
    "TexGenfv",
    "TexGeni",
    "TexGeniv",
    "FeedbackBuffer",
    "SelectBuffer",
    "RenderMode",
    "InitNames",
    "LoadName",
    "PassThrough",
    "PopName",
    "PushName",
    "DrawBuffer",
    "Clear",
    "ClearAccum",
    "ClearIndex",
    "ClearColor",
    "ClearStencil",
    "ClearDepth",
    "StencilMask",
    "ColorMask",
    "DepthMask",
    "IndexMask",
    "Accum",
    "Disable",
    "Enable",
    "Finish",
    "Flush",
    "PopAttrib",
    "PushAttrib",
    "Map1d",
    "Map1f",
    "Map2d",
    "Map2f",
    "MapGrid1d",
    "MapGrid1f",
    "MapGrid2d",
    "MapGrid2f",
    "EvalCoord1d",
    "EvalCoord1dv",
    "EvalCoord1f",
    "EvalCoord1fv",
    "EvalCoord2d",
    "EvalCoord2dv",
    "EvalCoord2f",
    "EvalCoord2fv",
    "EvalMesh1",
    "EvalPoint1",
    "EvalMesh2",
    "EvalPoint2",
    "AlphaFunc",
    "BlendFunc",
    "LogicOp",
    "StencilFunc",
    "StencilOp",
    "DepthFunc",
    "PixelZoom",
    "PixelTransferf",
    "PixelTransferi",
    "PixelStoref",
    "PixelStorei",
    "PixelMapfv",
    "PixelMapuiv",
    "PixelMapusv",
    "ReadBuffer",
    "CopyPixels",
    "ReadPixels",
    "DrawPixels",
    "GetBooleanv",
    "GetClipPlane",
    "GetDoublev",
    "GetError",
    "GetFloatv",
    "GetIntegerv",
    "GetLightfv",
    "GetLightiv",
    "GetMapdv",
    "GetMapfv",
    "GetMapiv",
    "GetMaterialfv",
    "GetMaterialiv",
    "GetPixelMapfv",
    "GetPixelMapuiv",
    "GetPixelMapusv",
    "GetPolygonStipple",
    "GetString",
    "GetTexEnvfv",
    "GetTexEnviv",
    "GetTexGendv",
    "GetTexGenfv",
    "GetTexGeniv",
    "GetTexImage",
    "GetTexParameterfv",
    "GetTexParameteriv",
    "GetTexLevelParameterfv",
    "GetTexLevelParameteriv",
    "IsEnabled",
    "IsList",
    "DepthRange",
    "Frustum",
    "LoadIdentity",
    "LoadMatrixf",
    "LoadMatrixd",
    "MatrixMode",
    "MultMatrixf",
    "MultMatrixd",
    "Ortho",
    "PopMatrix",
    "PushMatrix",
    "Rotated",
    "Rotatef",
    "Scaled",
    "Scalef",
    "Translated",
    "Translatef",
    "Viewport",
    "ArrayElement",
    "BindTexture",
    "ColorPointer",
    "DisableClientState",
    "DrawArrays",
    "DrawElements",
    "EdgeFlagPointer",
    "EnableClientState",
    "IndexPointer",
    "Indexub",
    "Indexubv",
    "InterleavedArrays",
    "NormalPointer",
    "PolygonOffset",
    "TexCoordPointer",
    "VertexPointer",
    "AreTexturesResident",
    "CopyTexImage1D",
    "CopyTexImage2D",
    "CopyTexSubImage1D",
    "CopyTexSubImage2D",
    "DeleteTextures",
    "GenTextures",
    "GetPointerv",
    "IsTexture",
    "PrioritizeTextures",
    "TexSubImage1D",
    "TexSubImage2D",
    "PopClientAttrib",
    "PushClientAttrib",
    "BlendColor",
    "BlendEquation",
    "DrawRangeElements",
    "TexImage3D",
    "TexSubImage3D",
    "CopyTexSubImage3D",
    "ActiveTexture",
    "ClientActiveTexture",
    "MultiTexCoord1d",
    "MultiTexCoord1dv",
    "MultiTexCoord1fARB",
    "MultiTexCoord1fvARB",
    "MultiTexCoord1i",
    "MultiTexCoord1iv",
    "MultiTexCoord1s",
    "MultiTexCoord1sv",
    "MultiTexCoord2d",
    "MultiTexCoord2dv",
    "MultiTexCoord2fARB",
    "MultiTexCoord2fvARB",
    "MultiTexCoord2i",
    "MultiTexCoord2iv",
    "MultiTexCoord2s",
    "MultiTexCoord2sv",
    "MultiTexCoord3d",
    "MultiTexCoord3dv",
    "MultiTexCoord3fARB",
    "MultiTexCoord3fvARB",
    "MultiTexCoord3i",
    "MultiTexCoord3iv",
    "MultiTexCoord3s",
    "MultiTexCoord3sv",
    "MultiTexCoord4d",
    "MultiTexCoord4dv",
    "MultiTexCoord4fARB",
    "MultiTexCoord4fvARB",
    "MultiTexCoord4i",
    "MultiTexCoord4iv",
    "MultiTexCoord4s",
    "MultiTexCoord4sv",
    "CompressedTexImage1D",
    "CompressedTexImage2D",
    "CompressedTexImage3D",
    "CompressedTexSubImage1D",
    "CompressedTexSubImage2D",
    "CompressedTexSubImage3D",
    "GetCompressedTexImage",
    "LoadTransposeMatrixd",
    "LoadTransposeMatrixf",
    "MultTransposeMatrixd",
    "MultTransposeMatrixf",
    "SampleCoverage",
    "BlendFuncSeparate",
    "FogCoordPointer",
    "FogCoordd",
    "FogCoorddv",
    "MultiDrawArrays",
    "PointParameterf",
    "PointParameterfv",
    "PointParameteri",
    "PointParameteriv",
    "SecondaryColor3b",
    "SecondaryColor3bv",
    "SecondaryColor3d",
    "SecondaryColor3dv",
    "SecondaryColor3i",
    "SecondaryColor3iv",
    "SecondaryColor3s",
    "SecondaryColor3sv",
    "SecondaryColor3ub",
    "SecondaryColor3ubv",
    "SecondaryColor3ui",
    "SecondaryColor3uiv",
    "SecondaryColor3us",
    "SecondaryColor3usv",
    "SecondaryColorPointer",
    "WindowPos2d",
    "WindowPos2dv",
    "WindowPos2f",
    "WindowPos2fv",
    "WindowPos2i",
    "WindowPos2iv",
    "WindowPos2s",
    "WindowPos2sv",
    "WindowPos3d",
    "WindowPos3dv",
    "WindowPos3f",
    "WindowPos3fv",
    "WindowPos3i",
    "WindowPos3iv",
    "WindowPos3s",
    "WindowPos3sv",
    "BeginQuery",
    "BindBuffer",
    "BufferData",
    "BufferSubData",
    "DeleteBuffers",
    "DeleteQueries",
    "EndQuery",
    "GenBuffers",
    "GenQueries",
    "GetBufferParameteriv",
    "GetBufferPointerv",
    "GetBufferSubData",
    "GetQueryObjectiv",
    "GetQueryObjectuiv",
    "GetQueryiv",
    "IsBuffer",
    "IsQuery",
    "MapBuffer",
    "UnmapBuffer",
    "AttachShader",
    "BindAttribLocation",
    "BlendEquationSeparate",
    "CompileShader",
    "CreateProgram",
    "CreateShader",
    "DeleteProgram",
    "DeleteShader",
    "DetachShader",
    "DisableVertexAttribArray",
    "DrawBuffers",
    "EnableVertexAttribArray",
    "GetActiveAttrib",
    "GetActiveUniform",
    "GetAttachedShaders",
    "GetAttribLocation",
    "GetProgramInfoLog",
    "GetProgramiv",
    "GetShaderInfoLog",
    "GetShaderSource",
    "GetShaderiv",
    "GetUniformLocation",
    "GetUniformfv",
    "GetUniformiv",
    "GetVertexAttribPointerv",
    "GetVertexAttribdv",
    "GetVertexAttribfv",
    "GetVertexAttribiv",
    "IsProgram",
    "IsShader",
    "LinkProgram",
    "ShaderSource",
    "StencilFuncSeparate",
    "StencilMaskSeparate",
    "StencilOpSeparate",
    "Uniform1f",
    "Uniform1fv",
    "Uniform1i",
    "Uniform1iv",
    "Uniform2f",
    "Uniform2fv",
    "Uniform2i",
    "Uniform2iv",
    "Uniform3f",
    "Uniform3fv",
    "Uniform3i",
    "Uniform3iv",
    "Uniform4f",
    "Uniform4fv",
    "Uniform4i",
    "Uniform4iv",
    "UniformMatrix2fv",
    "UniformMatrix3fv",
    "UniformMatrix4fv",
    "UseProgram",
    "ValidateProgram",
    "VertexAttrib1d",
    "VertexAttrib1dv",
    "VertexAttrib1s",
    "VertexAttrib1sv",
    "VertexAttrib2d",
    "VertexAttrib2dv",
    "VertexAttrib2s",
    "VertexAttrib2sv",
    "VertexAttrib3d",
    "VertexAttrib3dv",
    "VertexAttrib3s",
    "VertexAttrib3sv",
    "VertexAttrib4Nbv",
    "VertexAttrib4Niv",
    "VertexAttrib4Nsv",
    "VertexAttrib4Nub",
    "VertexAttrib4Nubv",
    "VertexAttrib4Nuiv",
    "VertexAttrib4Nusv",
    "VertexAttrib4bv",
    "VertexAttrib4d",
    "VertexAttrib4dv",
    "VertexAttrib4iv",
    "VertexAttrib4s",
    "VertexAttrib4sv",
    "VertexAttrib4ubv",
    "VertexAttrib4uiv",
    "VertexAttrib4usv",
    "VertexAttribPointer",
    "UniformMatrix2x3fv",
    "UniformMatrix2x4fv",
    "UniformMatrix3x2fv",
    "UniformMatrix3x4fv",
    "UniformMatrix4x2fv",
    "UniformMatrix4x3fv",
    "BeginConditionalRender",
    "BeginTransformFeedback",
    "BindBufferBase",
    "BindBufferRange",
    "BindFragDataLocation",
    "ClampColor",
    "ClearBufferfi",
    "ClearBufferfv",
    "ClearBufferiv",
    "ClearBufferuiv",
    "ColorMaski",
    "Disablei",
    "Enablei",
    "EndConditionalRender",
    "EndTransformFeedback",
    "GetBooleani_v",
    "GetFragDataLocation",
    "GetIntegeri_v",
    "GetStringi",
    "GetTexParameterIiv",
    "GetTexParameterIuiv",
    "GetTransformFeedbackVarying",
    "GetUniformuiv",
    "GetVertexAttribIiv",
    "GetVertexAttribIuiv",
    "IsEnabledi",
    "TexParameterIiv",
    "TexParameterIuiv",
    "TransformFeedbackVaryings",
    "Uniform1ui",
    "Uniform1uiv",
    "Uniform2ui",
    "Uniform2uiv",
    "Uniform3ui",
    "Uniform3uiv",
    "Uniform4ui",
    "Uniform4uiv",
    "VertexAttribI1iv",
    "VertexAttribI1uiv",
    "VertexAttribI4bv",
    "VertexAttribI4sv",
    "VertexAttribI4ubv",
    "VertexAttribI4usv",
    "VertexAttribIPointer",
    "PrimitiveRestartIndex",
    "TexBuffer",
    "FramebufferTexture",
    "GetBufferParameteri64v",
    "GetInteger64i_v",
    "VertexAttribDivisor",
    "MinSampleShading",
    "MemoryBarrierByRegion",
    "BindProgramARB",
    "DeleteProgramsARB",
    "GenProgramsARB",
    "GetProgramEnvParameterdvARB",
    "GetProgramEnvParameterfvARB",
    "GetProgramLocalParameterdvARB",
    "GetProgramLocalParameterfvARB",
    "GetProgramStringARB",
    "GetProgramivARB",
    "IsProgramARB",
    "ProgramEnvParameter4dARB",
    "ProgramEnvParameter4dvARB",
    "ProgramEnvParameter4fARB",
    "ProgramEnvParameter4fvARB",
    "ProgramLocalParameter4dARB",
    "ProgramLocalParameter4dvARB",
    "ProgramLocalParameter4fARB",
    "ProgramLocalParameter4fvARB",
    "ProgramStringARB",
    "VertexAttrib1fARB",
    "VertexAttrib1fvARB",
    "VertexAttrib2fARB",
    "VertexAttrib2fvARB",
    "VertexAttrib3fARB",
    "VertexAttrib3fvARB",
    "VertexAttrib4fARB",
    "VertexAttrib4fvARB",
    "AttachObjectARB",
    "CreateProgramObjectARB",
    "CreateShaderObjectARB",
    "DeleteObjectARB",
    "DetachObjectARB",
    "GetAttachedObjectsARB",
    "GetHandleARB",
    "GetInfoLogARB",
    "GetObjectParameterfvARB",
    "GetObjectParameterivARB",
    "DrawArraysInstanced",
    "DrawElementsInstanced",
    "BindFramebuffer",
    "BindRenderbuffer",
    "BlitFramebuffer",
    "CheckFramebufferStatus",
    "DeleteFramebuffers",
    "DeleteRenderbuffers",
    "FramebufferRenderbuffer",
    "FramebufferTexture1D",
    "FramebufferTexture2D",
    "FramebufferTexture3D",
    "FramebufferTextureLayer",
    "GenFramebuffers",
    "GenRenderbuffers",
    "GenerateMipmap",
    "GetFramebufferAttachmentParameteriv",
    "GetRenderbufferParameteriv",
    "IsFramebuffer",
    "IsRenderbuffer",
    "RenderbufferStorage",
    "RenderbufferStorageMultisample",
    "FlushMappedBufferRange",
    "MapBufferRange",
    "BindVertexArray",
    "DeleteVertexArrays",
    "GenVertexArrays",
    "IsVertexArray",
    "GetActiveUniformBlockName",
    "GetActiveUniformBlockiv",
    "GetActiveUniformName",
    "GetActiveUniformsiv",
    "GetUniformBlockIndex",
    "GetUniformIndices",
    "UniformBlockBinding",
    "CopyBufferSubData",
    "ClientWaitSync",
    "DeleteSync",
    "FenceSync",
    "GetInteger64v",
    "GetSynciv",
    "IsSync",
    "WaitSync",
    "DrawElementsBaseVertex",
    "DrawElementsInstancedBaseVertex",
    "DrawRangeElementsBaseVertex",
    "MultiDrawElementsBaseVertex",
    "ProvokingVertex",
    "GetMultisamplefv",
    "SampleMaski",
    "TexImage2DMultisample",
    "TexImage3DMultisample",
    "BlendEquationSeparateiARB",
    "BlendEquationiARB",
    "BlendFuncSeparateiARB",
    "BlendFunciARB",
    "BindFragDataLocationIndexed",
    "GetFragDataIndex",
    "BindSampler",
    "DeleteSamplers",
    "GenSamplers",
    "GetSamplerParameterIiv",
    "GetSamplerParameterIuiv",
    "GetSamplerParameterfv",
    "GetSamplerParameteriv",
    "IsSampler",
    "SamplerParameterIiv",
    "SamplerParameterIuiv",
    "SamplerParameterf",
    "SamplerParameterfv",
    "SamplerParameteri",
    "SamplerParameteriv",
    "GetQueryObjecti64v",
    "GetQueryObjectui64v",
    "QueryCounter",
    "ColorP3ui",
    "ColorP3uiv",
    "ColorP4ui",
    "ColorP4uiv",
    "MultiTexCoordP1ui",
    "MultiTexCoordP1uiv",
    "MultiTexCoordP2ui",
    "MultiTexCoordP2uiv",
    "MultiTexCoordP3ui",
    "MultiTexCoordP3uiv",
    "MultiTexCoordP4ui",
    "MultiTexCoordP4uiv",
    "NormalP3ui",
    "NormalP3uiv",
    "SecondaryColorP3ui",
    "SecondaryColorP3uiv",
    "TexCoordP1ui",
    "TexCoordP1uiv",
    "TexCoordP2ui",
    "TexCoordP2uiv",
    "TexCoordP3ui",
    "TexCoordP3uiv",
    "TexCoordP4ui",
    "TexCoordP4uiv",
    "VertexAttribP1ui",
    "VertexAttribP1uiv",
    "VertexAttribP2ui",
    "VertexAttribP2uiv",
    "VertexAttribP3ui",
    "VertexAttribP3uiv",
    "VertexAttribP4ui",
    "VertexAttribP4uiv",
    "VertexP2ui",
    "VertexP2uiv",
    "VertexP3ui",
    "VertexP3uiv",
    "VertexP4ui",
    "VertexP4uiv",
    "DrawArraysIndirect",
    "DrawElementsIndirect",
    "GetUniformdv",
    "Uniform1d",
    "Uniform1dv",
    "Uniform2d",
    "Uniform2dv",
    "Uniform3d",
    "Uniform3dv",
    "Uniform4d",
    "Uniform4dv",
    "UniformMatrix2dv",
    "UniformMatrix2x3dv",
    "UniformMatrix2x4dv",
    "UniformMatrix3dv",
    "UniformMatrix3x2dv",
    "UniformMatrix3x4dv",
    "UniformMatrix4dv",
    "UniformMatrix4x2dv",
    "UniformMatrix4x3dv",
    "GetActiveSubroutineName",
    "GetActiveSubroutineUniformName",
    "GetActiveSubroutineUniformiv",
    "GetProgramStageiv",
    "GetSubroutineIndex",
    "GetSubroutineUniformLocation",
    "GetUniformSubroutineuiv",
    "UniformSubroutinesuiv",
    "PatchParameterfv",
    "PatchParameteri",
    "BindTransformFeedback",
    "DeleteTransformFeedbacks",
    "DrawTransformFeedback",
    "GenTransformFeedbacks",
    "IsTransformFeedback",
    "PauseTransformFeedback",
    "ResumeTransformFeedback",
    "BeginQueryIndexed",
    "DrawTransformFeedbackStream",
    "EndQueryIndexed",
    "GetQueryIndexediv",
    "ClearDepthf",
    "DepthRangef",
    "GetShaderPrecisionFormat",
    "ReleaseShaderCompiler",
    "ShaderBinary",
    "GetProgramBinary",
    "ProgramBinary",
    "ProgramParameteri",
    "GetVertexAttribLdv",
    "VertexAttribL1d",
    "VertexAttribL1dv",
    "VertexAttribL2d",
    "VertexAttribL2dv",
    "VertexAttribL3d",
    "VertexAttribL3dv",
    "VertexAttribL4d",
    "VertexAttribL4dv",
    "VertexAttribLPointer",
    "DepthRangeArrayv",
    "DepthRangeIndexed",
    "GetDoublei_v",
    "GetFloati_v",
    "ScissorArrayv",
    "ScissorIndexed",
    "ScissorIndexedv",
    "ViewportArrayv",
    "ViewportIndexedf",
    "ViewportIndexedfv",
    "GetGraphicsResetStatusARB",
    "GetnCompressedTexImageARB",
    "GetnMapdvARB",
    "GetnMapfvARB",
    "GetnMapivARB",
    "GetnPixelMapfvARB",
    "GetnPixelMapuivARB",
    "GetnPixelMapusvARB",
    "GetnPolygonStippleARB",
    "GetnTexImageARB",
    "GetnUniformdvARB",
    "GetnUniformfvARB",
    "GetnUniformivARB",
    "GetnUniformuivARB",
    "ReadnPixelsARB",
    "DrawArraysInstancedBaseInstance",
    "DrawElementsInstancedBaseInstance",
    "DrawElementsInstancedBaseVertexBaseInstance",
    "DrawTransformFeedbackInstanced",
    "DrawTransformFeedbackStreamInstanced",
    "GetInternalformativ",
    "GetActiveAtomicCounterBufferiv",
    "BindImageTexture",
    "MemoryBarrier",
    "TexStorage1D",
    "TexStorage2D",
    "TexStorage3D",
    "TextureStorage1DEXT",
    "TextureStorage2DEXT",
    "TextureStorage3DEXT",
    "ClearBufferData",
    "ClearBufferSubData",
    "DispatchCompute",
    "DispatchComputeIndirect",
    "CopyImageSubData",
    "TextureView",
    "BindVertexBuffer",
    "VertexAttribBinding",
    "VertexAttribFormat",
    "VertexAttribIFormat",
    "VertexAttribLFormat",
    "VertexBindingDivisor",
    "FramebufferParameteri",
    "GetFramebufferParameteriv",
    "GetInternalformati64v",
    "MultiDrawArraysIndirect",
    "MultiDrawElementsIndirect",
    "GetProgramInterfaceiv",
    "GetProgramResourceIndex",
    "GetProgramResourceLocation",
    "GetProgramResourceLocationIndex",
    "GetProgramResourceName",
    "GetProgramResourceiv",
    "ShaderStorageBlockBinding",
    "TexBufferRange",
    "TexStorage2DMultisample",
    "TexStorage3DMultisample",
    "BufferStorage",
    "ClearTexImage",
    "ClearTexSubImage",
    "BindBuffersBase",
    "BindBuffersRange",
    "BindImageTextures",
    "BindSamplers",
    "BindTextures",
    "BindVertexBuffers",
    "GetImageHandleARB",
    "GetTextureHandleARB",
    "GetTextureSamplerHandleARB",
    "GetVertexAttribLui64vARB",
    "IsImageHandleResidentARB",
    "IsTextureHandleResidentARB",
    "MakeImageHandleNonResidentARB",
    "MakeImageHandleResidentARB",
    "MakeTextureHandleNonResidentARB",
    "MakeTextureHandleResidentARB",
    "ProgramUniformHandleui64ARB",
    "ProgramUniformHandleui64vARB",
    "UniformHandleui64ARB",
    "UniformHandleui64vARB",
    "VertexAttribL1ui64ARB",
    "VertexAttribL1ui64vARB",
    "DispatchComputeGroupSizeARB",
    "MultiDrawArraysIndirectCountARB",
    "MultiDrawElementsIndirectCountARB",
    "ClipControl",
    "BindTextureUnit",
    "BlitNamedFramebuffer",
    "CheckNamedFramebufferStatus",
    "ClearNamedBufferData",
    "ClearNamedBufferSubData",
    "ClearNamedFramebufferfi",
    "ClearNamedFramebufferfv",
    "ClearNamedFramebufferiv",
    "ClearNamedFramebufferuiv",
    "CompressedTextureSubImage1D",
    "CompressedTextureSubImage2D",
    "CompressedTextureSubImage3D",
    "CopyNamedBufferSubData",
    "CopyTextureSubImage1D",
    "CopyTextureSubImage2D",
    "CopyTextureSubImage3D",
    "CreateBuffers",
    "CreateFramebuffers",
    "CreateProgramPipelines",
    "CreateQueries",
    "CreateRenderbuffers",
    "CreateSamplers",
    "CreateTextures",
    "CreateTransformFeedbacks",
    "CreateVertexArrays",
    "DisableVertexArrayAttrib",
    "EnableVertexArrayAttrib",
    "FlushMappedNamedBufferRange",
    "GenerateTextureMipmap",
    "GetCompressedTextureImage",
    "GetNamedBufferParameteri64v",
    "GetNamedBufferParameteriv",
    "GetNamedBufferPointerv",
    "GetNamedBufferSubData",
    "GetNamedFramebufferAttachmentParameteriv",
    "GetNamedFramebufferParameteriv",
    "GetNamedRenderbufferParameteriv",
    "GetQueryBufferObjecti64v",
    "GetQueryBufferObjectiv",
    "GetQueryBufferObjectui64v",
    "GetQueryBufferObjectuiv",
    "GetTextureImage",
    "GetTextureLevelParameterfv",
    "GetTextureLevelParameteriv",
    "GetTextureParameterIiv",
    "GetTextureParameterIuiv",
    "GetTextureParameterfv",
    "GetTextureParameteriv",
    "GetTransformFeedbacki64_v",
    "GetTransformFeedbacki_v",
    "GetTransformFeedbackiv",
    "GetVertexArrayIndexed64iv",
    "GetVertexArrayIndexediv",
    "GetVertexArrayiv",
    "InvalidateNamedFramebufferData",
    "InvalidateNamedFramebufferSubData",
    "MapNamedBuffer",
    "MapNamedBufferRange",
    "NamedBufferData",
    "NamedBufferStorage",
    "NamedBufferSubData",
    "NamedFramebufferDrawBuffer",
    "NamedFramebufferDrawBuffers",
    "NamedFramebufferParameteri",
    "NamedFramebufferReadBuffer",
    "NamedFramebufferRenderbuffer",
    "NamedFramebufferTexture",
    "NamedFramebufferTextureLayer",
    "NamedRenderbufferStorage",
    "NamedRenderbufferStorageMultisample",
    "TextureBuffer",
    "TextureBufferRange",
    "TextureParameterIiv",
    "TextureParameterIuiv",
    "TextureParameterf",
    "TextureParameterfv",
    "TextureParameteri",
    "TextureParameteriv",
    "TextureStorage1D",
    "TextureStorage2D",
    "TextureStorage2DMultisample",
    "TextureStorage3D",
    "TextureStorage3DMultisample",
    "TextureSubImage1D",
    "TextureSubImage2D",
    "TextureSubImage3D",
    "TransformFeedbackBufferBase",
    "TransformFeedbackBufferRange",
    "UnmapNamedBufferEXT",
    "VertexArrayAttribBinding",
    "VertexArrayAttribFormat",
    "VertexArrayAttribIFormat",
    "VertexArrayAttribLFormat",
    "VertexArrayBindingDivisor",
    "VertexArrayElementBuffer",
    "VertexArrayVertexBuffer",
    "VertexArrayVertexBuffers",
    "GetCompressedTextureSubImage",
    "GetTextureSubImage",
    "BufferPageCommitmentARB",
    "NamedBufferPageCommitmentARB",
    "GetUniformi64vARB",
    "GetUniformui64vARB",
    "GetnUniformi64vARB",
    "GetnUniformui64vARB",
    "ProgramUniform1i64ARB",
    "ProgramUniform1i64vARB",
    "ProgramUniform1ui64ARB",
    "ProgramUniform1ui64vARB",
    "ProgramUniform2i64ARB",
    "ProgramUniform2i64vARB",
    "ProgramUniform2ui64ARB",
    "ProgramUniform2ui64vARB",
    "ProgramUniform3i64ARB",
    "ProgramUniform3i64vARB",
    "ProgramUniform3ui64ARB",
    "ProgramUniform3ui64vARB",
    "ProgramUniform4i64ARB",
    "ProgramUniform4i64vARB",
    "ProgramUniform4ui64ARB",
    "ProgramUniform4ui64vARB",
    "Uniform1i64ARB",
    "Uniform1i64vARB",
    "Uniform1ui64ARB",
    "Uniform1ui64vARB",
    "Uniform2i64ARB",
    "Uniform2i64vARB",
    "Uniform2ui64ARB",
    "Uniform2ui64vARB",
    "Uniform3i64ARB",
    "Uniform3i64vARB",
    "Uniform3ui64ARB",
    "Uniform3ui64vARB",
    "Uniform4i64ARB",
    "Uniform4i64vARB",
    "Uniform4ui64ARB",
    "Uniform4ui64vARB",
    "EvaluateDepthValuesARB",
    "FramebufferSampleLocationsfvARB",
    "NamedFramebufferSampleLocationsfvARB",
    "SpecializeShaderARB",
    "InvalidateBufferData",
    "InvalidateBufferSubData",
    "InvalidateFramebuffer",
    "InvalidateSubFramebuffer",
    "InvalidateTexImage",
    "InvalidateTexSubImage",
    "DrawTexfOES",
    "DrawTexfvOES",
    "DrawTexiOES",
    "DrawTexivOES",
    "DrawTexsOES",
    "DrawTexsvOES",
    "DrawTexxOES",
    "DrawTexxvOES",
    "PointSizePointerOES",
    "QueryMatrixxOES",
    "ColorPointerEXT",
    "EdgeFlagPointerEXT",
    "IndexPointerEXT",
    "NormalPointerEXT",
    "TexCoordPointerEXT",
    "VertexPointerEXT",
    "DiscardFramebufferEXT",
    "ActiveShaderProgram",
    "BindProgramPipeline",
    "CreateShaderProgramv",
    "DeleteProgramPipelines",
    "GenProgramPipelines",
    "GetProgramPipelineInfoLog",
    "GetProgramPipelineiv",
    "IsProgramPipeline",
    "LockArraysEXT",
    "ProgramUniform1d",
    "ProgramUniform1dv",
    "ProgramUniform1f",
    "ProgramUniform1fv",
    "ProgramUniform1i",
    "ProgramUniform1iv",
    "ProgramUniform1ui",
    "ProgramUniform1uiv",
    "ProgramUniform2d",
    "ProgramUniform2dv",
    "ProgramUniform2f",
    "ProgramUniform2fv",
    "ProgramUniform2i",
    "ProgramUniform2iv",
    "ProgramUniform2ui",
    "ProgramUniform2uiv",
    "ProgramUniform3d",
    "ProgramUniform3dv",
    "ProgramUniform3f",
    "ProgramUniform3fv",
    "ProgramUniform3i",
    "ProgramUniform3iv",
    "ProgramUniform3ui",
    "ProgramUniform3uiv",
    "ProgramUniform4d",
    "ProgramUniform4dv",
    "ProgramUniform4f",
    "ProgramUniform4fv",
    "ProgramUniform4i",
    "ProgramUniform4iv",
    "ProgramUniform4ui",
    "ProgramUniform4uiv",
    "ProgramUniformMatrix2dv",
    "ProgramUniformMatrix2fv",
    "ProgramUniformMatrix2x3dv",
    "ProgramUniformMatrix2x3fv",
    "ProgramUniformMatrix2x4dv",
    "ProgramUniformMatrix2x4fv",
    "ProgramUniformMatrix3dv",
    "ProgramUniformMatrix3fv",
    "ProgramUniformMatrix3x2dv",
    "ProgramUniformMatrix3x2fv",
    "ProgramUniformMatrix3x4dv",
    "ProgramUniformMatrix3x4fv",
    "ProgramUniformMatrix4dv",
    "ProgramUniformMatrix4fv",
    "ProgramUniformMatrix4x2dv",
    "ProgramUniformMatrix4x2fv",
    "ProgramUniformMatrix4x3dv",
    "ProgramUniformMatrix4x3fv",
    "UnlockArraysEXT",
    "UseProgramStages",
    "ValidateProgramPipeline",
    "FramebufferTexture2DMultisampleEXT",
    "DebugMessageCallback",
    "DebugMessageControl",
    "DebugMessageInsert",
    "GetDebugMessageLog",
    "GetObjectLabel",
    "GetObjectPtrLabel",
    "ObjectLabel",
    "ObjectPtrLabel",
    "PopDebugGroup",
    "PushDebugGroup",
    "SecondaryColor3fEXT",
    "SecondaryColor3fvEXT",
    "MultiDrawElements",
    "FogCoordfEXT",
    "FogCoordfvEXT",
    "WindowPos4dMESA",
    "WindowPos4dvMESA",
    "WindowPos4fMESA",
    "WindowPos4fvMESA",
    "WindowPos4iMESA",
    "WindowPos4ivMESA",
    "WindowPos4sMESA",
    "WindowPos4svMESA",
    "MultiModeDrawArraysIBM",
    "MultiModeDrawElementsIBM",
    "VertexAttrib1dNV",
    "VertexAttrib1dvNV",
    "VertexAttrib1fNV",
    "VertexAttrib1fvNV",
    "VertexAttrib1sNV",
    "VertexAttrib1svNV",
    "VertexAttrib2dNV",
    "VertexAttrib2dvNV",
    "VertexAttrib2fNV",
    "VertexAttrib2fvNV",
    "VertexAttrib2sNV",
    "VertexAttrib2svNV",
    "VertexAttrib3dNV",
    "VertexAttrib3dvNV",
    "VertexAttrib3fNV",
    "VertexAttrib3fvNV",
    "VertexAttrib3sNV",
    "VertexAttrib3svNV",
    "VertexAttrib4dNV",
    "VertexAttrib4dvNV",
    "VertexAttrib4fNV",
    "VertexAttrib4fvNV",
    "VertexAttrib4sNV",
    "VertexAttrib4svNV",
    "VertexAttrib4ubNV",
    "VertexAttrib4ubvNV",
    "VertexAttribs1dvNV",
    "VertexAttribs1fvNV",
    "VertexAttribs1svNV",
    "VertexAttribs2dvNV",
    "VertexAttribs2fvNV",
    "VertexAttribs2svNV",
    "VertexAttribs3dvNV",
    "VertexAttribs3fvNV",
    "VertexAttribs3svNV",
    "VertexAttribs4dvNV",
    "VertexAttribs4fvNV",
    "VertexAttribs4svNV",
    "VertexAttribs4ubvNV",
    "AlphaFragmentOp1ATI",
    "AlphaFragmentOp2ATI",
    "AlphaFragmentOp3ATI",
    "BeginFragmentShaderATI",
    "BindFragmentShaderATI",
    "ColorFragmentOp1ATI",
    "ColorFragmentOp2ATI",
    "ColorFragmentOp3ATI",
    "DeleteFragmentShaderATI",
    "EndFragmentShaderATI",
    "GenFragmentShadersATI",
    "PassTexCoordATI",
    "SampleMapATI",
    "SetFragmentShaderConstantATI",
    "DepthRangeArrayfvOES",
    "DepthRangeIndexedfOES",
    "ActiveStencilFaceEXT",
    "PrimitiveRestartNV",
    "GetTexGenxvOES",
    "TexGenxOES",
    "TexGenxvOES",
    "DepthBoundsEXT",
    "BindFramebufferEXT",
    "BindRenderbufferEXT",
    "StringMarkerGREMEDY",
    "VertexAttribI1iEXT",
    "VertexAttribI1uiEXT",
    "VertexAttribI2iEXT",
    "VertexAttribI2ivEXT",
    "VertexAttribI2uiEXT",
    "VertexAttribI2uivEXT",
    "VertexAttribI3iEXT",
    "VertexAttribI3ivEXT",
    "VertexAttribI3uiEXT",
    "VertexAttribI3uivEXT",
    "VertexAttribI4iEXT",
    "VertexAttribI4ivEXT",
    "VertexAttribI4uiEXT",
    "VertexAttribI4uivEXT",
    "ClearColorIiEXT",
    "ClearColorIuiEXT",
    "BindBufferOffsetEXT",
    "BeginPerfMonitorAMD",
    "DeletePerfMonitorsAMD",
    "EndPerfMonitorAMD",
    "GenPerfMonitorsAMD",
    "GetPerfMonitorCounterDataAMD",
    "GetPerfMonitorCounterInfoAMD",
    "GetPerfMonitorCounterStringAMD",
    "GetPerfMonitorCountersAMD",
    "GetPerfMonitorGroupStringAMD",
    "GetPerfMonitorGroupsAMD",
    "SelectPerfMonitorCountersAMD",
    "TextureBarrierNV",
    "BeginPerfQueryINTEL",
    "CreatePerfQueryINTEL",
    "DeletePerfQueryINTEL",
    "EndPerfQueryINTEL",
    "GetFirstPerfQueryIdINTEL",
    "GetNextPerfQueryIdINTEL",
    "GetPerfCounterInfoINTEL",
    "GetPerfQueryDataINTEL",
    "GetPerfQueryIdByNameINTEL",
    "GetPerfQueryInfoINTEL",
    "PolygonOffsetClampEXT",
    "SubpixelPrecisionBiasNV",
    "ConservativeRasterParameterfNV",
    "ConservativeRasterParameteriNV",
    "WindowRectanglesEXT",
    "BufferStorageMemEXT",
    "CreateMemoryObjectsEXT",
    "DeleteMemoryObjectsEXT",
    "DeleteSemaphoresEXT",
    "GenSemaphoresEXT",
    "GetMemoryObjectParameterivEXT",
    "GetSemaphoreParameterui64vEXT",
    "GetUnsignedBytei_vEXT",
    "GetUnsignedBytevEXT",
    "IsMemoryObjectEXT",
    "IsSemaphoreEXT",
    "MemoryObjectParameterivEXT",
    "NamedBufferStorageMemEXT",
    "SemaphoreParameterui64vEXT",
    "SignalSemaphoreEXT",
    "TexStorageMem1DEXT",
    "TexStorageMem2DEXT",
    "TexStorageMem2DMultisampleEXT",
    "TexStorageMem3DEXT",
    "TexStorageMem3DMultisampleEXT",
    "TextureStorageMem1DEXT",
    "TextureStorageMem2DEXT",
    "TextureStorageMem2DMultisampleEXT",
    "TextureStorageMem3DEXT",
    "TextureStorageMem3DMultisampleEXT",
    "WaitSemaphoreEXT",
    "ImportMemoryFdEXT",
    "ImportSemaphoreFdEXT",
    "FramebufferFetchBarrierEXT",
    "NamedRenderbufferStorageMultisampleAdvancedAMD",
    "RenderbufferStorageMultisampleAdvancedAMD",
    "StencilFuncSeparateATI",
    "ProgramEnvParameters4fvEXT",
    "ProgramLocalParameters4fvEXT",
    "EGLImageTargetRenderbufferStorageOES",
    "EGLImageTargetTexture2DOES",
    "AlphaFuncx",
    "ClearColorx",
    "ClearDepthx",
    "Color4x",
    "DepthRangex",
    "Fogx",
    "Fogxv",
    "Frustumf",
    "Frustumx",
    "LightModelx",
    "LightModelxv",
    "Lightx",
    "Lightxv",
    "LineWidthx",
    "LoadMatrixx",
    "Materialx",
    "Materialxv",
    "MultMatrixx",
    "MultiTexCoord4x",
    "Normal3x",
    "Orthof",
    "Orthox",
    "PointSizex",
    "PolygonOffsetx",
    "Rotatex",
    "SampleCoveragex",
    "Scalex",
    "TexEnvx",
    "TexEnvxv",
    "TexParameterx",
    "Translatex",
    "ClipPlanef",
    "ClipPlanex",
    "GetClipPlanef",
    "GetClipPlanex",
    "GetFixedv",
    "GetLightxv",
    "GetMaterialxv",
    "GetTexEnvxv",
    "GetTexParameterxv",
    "PointParameterx",
    "PointParameterxv",
    "TexParameterxv",
    "BlendBarrier",
    "PrimitiveBoundingBox",
    "MaxShaderCompilerThreadsKHR",
    "MatrixLoadfEXT",
    "MatrixLoaddEXT",
    "MatrixMultfEXT",
    "MatrixMultdEXT",
    "MatrixLoadIdentityEXT",
    "MatrixRotatefEXT",
    "MatrixRotatedEXT",
    "MatrixScalefEXT",
    "MatrixScaledEXT",
    "MatrixTranslatefEXT",
    "MatrixTranslatedEXT",
    "MatrixOrthoEXT",
    "MatrixFrustumEXT",
    "MatrixPushEXT",
    "MatrixPopEXT",
    "MatrixLoadTransposefEXT",
    "MatrixLoadTransposedEXT",
    "MatrixMultTransposefEXT",
    "MatrixMultTransposedEXT",
    "BindMultiTextureEXT",
    "NamedBufferDataEXT",
    "NamedBufferSubDataEXT",
    "NamedBufferStorageEXT",
    "MapNamedBufferRangeEXT",
    "TextureImage1DEXT",
    "TextureImage2DEXT",
    "TextureImage3DEXT",
    "TextureSubImage1DEXT",
    "TextureSubImage2DEXT",
    "TextureSubImage3DEXT",
    "CopyTextureImage1DEXT",
    "CopyTextureImage2DEXT",
    "CopyTextureSubImage1DEXT",
    "CopyTextureSubImage2DEXT",
    "CopyTextureSubImage3DEXT",
    "MapNamedBufferEXT",
    "GetTextureParameterivEXT",
    "GetTextureParameterfvEXT",
    "TextureParameteriEXT",
    "TextureParameterivEXT",
    "TextureParameterfEXT",
    "TextureParameterfvEXT",
    "GetTextureImageEXT",
    "GetTextureLevelParameterivEXT",
    "GetTextureLevelParameterfvEXT",
    "GetNamedBufferSubDataEXT",
    "GetNamedBufferPointervEXT",
    "GetNamedBufferParameterivEXT",
    "FlushMappedNamedBufferRangeEXT",
    "FramebufferDrawBufferEXT",
    "FramebufferDrawBuffersEXT",
    "FramebufferReadBufferEXT",
    "GetFramebufferParameterivEXT",
    "CheckNamedFramebufferStatusEXT",
    "NamedFramebufferTexture1DEXT",
    "NamedFramebufferTexture2DEXT",
    "NamedFramebufferTexture3DEXT",
    "NamedFramebufferRenderbufferEXT",
    "GetNamedFramebufferAttachmentParameterivEXT",
    "EnableClientStateiEXT",
    "DisableClientStateiEXT",
    "GetPointerIndexedvEXT",
    "MultiTexEnviEXT",
    "MultiTexEnvivEXT",
    "MultiTexEnvfEXT",
    "MultiTexEnvfvEXT",
    "GetMultiTexEnvivEXT",
    "GetMultiTexEnvfvEXT",
    "MultiTexParameteriEXT",
    "MultiTexParameterivEXT",
    "MultiTexParameterfEXT",
    "MultiTexParameterfvEXT",
    "GetMultiTexImageEXT",
    "MultiTexImage1DEXT",
    "MultiTexImage2DEXT",
    "MultiTexImage3DEXT",
    "MultiTexSubImage1DEXT",
    "MultiTexSubImage2DEXT",
    "MultiTexSubImage3DEXT",
    "GetMultiTexParameterivEXT",
    "GetMultiTexParameterfvEXT",
    "CopyMultiTexImage1DEXT",
    "CopyMultiTexImage2DEXT",
    "CopyMultiTexSubImage1DEXT",
    "CopyMultiTexSubImage2DEXT",
    "CopyMultiTexSubImage3DEXT",
    "MultiTexGendEXT",
    "MultiTexGendvEXT",
    "MultiTexGenfEXT",
    "MultiTexGenfvEXT",
    "MultiTexGeniEXT",
    "MultiTexGenivEXT",
    "GetMultiTexGendvEXT",
    "GetMultiTexGenfvEXT",
    "GetMultiTexGenivEXT",
    "MultiTexCoordPointerEXT",
    "BindImageTextureEXT",
    "CompressedTextureImage1DEXT",
    "CompressedTextureImage2DEXT",
    "CompressedTextureImage3DEXT",
    "CompressedTextureSubImage1DEXT",
    "CompressedTextureSubImage2DEXT",
    "CompressedTextureSubImage3DEXT",
    "GetCompressedTextureImageEXT",
    "CompressedMultiTexImage1DEXT",
    "CompressedMultiTexImage2DEXT",
    "CompressedMultiTexImage3DEXT",
    "CompressedMultiTexSubImage1DEXT",
    "CompressedMultiTexSubImage2DEXT",
    "CompressedMultiTexSubImage3DEXT",
    "GetCompressedMultiTexImageEXT",
    "GetMultiTexLevelParameterivEXT",
    "GetMultiTexLevelParameterfvEXT",
    "FramebufferParameteriMESA",
    "GetFramebufferParameterivMESA",
    "NamedRenderbufferStorageEXT",
    "GetNamedRenderbufferParameterivEXT",
    "ClientAttribDefaultEXT",
    "PushClientAttribDefaultEXT",
    "NamedProgramStringEXT",
    "GetNamedProgramStringEXT",
    "NamedProgramLocalParameter4fEXT",
    "NamedProgramLocalParameter4fvEXT",
    "GetNamedProgramLocalParameterfvEXT",
    "NamedProgramLocalParameter4dEXT",
    "NamedProgramLocalParameter4dvEXT",
    "GetNamedProgramLocalParameterdvEXT",
    "GetNamedProgramivEXT",
    "TextureBufferEXT",
    "MultiTexBufferEXT",
    "TextureParameterIivEXT",
    "TextureParameterIuivEXT",
    "GetTextureParameterIivEXT",
    "GetTextureParameterIuivEXT",
    "MultiTexParameterIivEXT",
    "MultiTexParameterIuivEXT",
    "GetMultiTexParameterIivEXT",
    "GetMultiTexParameterIuivEXT",
    "NamedProgramLocalParameters4fvEXT",
    "GenerateTextureMipmapEXT",
    "GenerateMultiTexMipmapEXT",
    "NamedRenderbufferStorageMultisampleEXT",
    "NamedCopyBufferSubDataEXT",
    "VertexArrayVertexOffsetEXT",
    "VertexArrayColorOffsetEXT",
    "VertexArrayEdgeFlagOffsetEXT",
    "VertexArrayIndexOffsetEXT",
    "VertexArrayNormalOffsetEXT",
    "VertexArrayTexCoordOffsetEXT",
    "VertexArrayMultiTexCoordOffsetEXT",
    "VertexArrayFogCoordOffsetEXT",
    "VertexArraySecondaryColorOffsetEXT",
    "VertexArrayVertexAttribOffsetEXT",
    "VertexArrayVertexAttribIOffsetEXT",
    "EnableVertexArrayEXT",
    "DisableVertexArrayEXT",
    "EnableVertexArrayAttribEXT",
    "DisableVertexArrayAttribEXT",
    "GetVertexArrayIntegervEXT",
    "GetVertexArrayPointervEXT",
    "GetVertexArrayIntegeri_vEXT",
    "GetVertexArrayPointeri_vEXT",
    "ClearNamedBufferDataEXT",
    "ClearNamedBufferSubDataEXT",
    "NamedFramebufferParameteriEXT",
    "GetNamedFramebufferParameterivEXT",
    "VertexArrayVertexAttribLOffsetEXT",
    "VertexArrayVertexAttribDivisorEXT",
    "TextureBufferRangeEXT",
    "TextureStorage2DMultisampleEXT",
    "TextureStorage3DMultisampleEXT",
    "VertexArrayBindVertexBufferEXT",
    "VertexArrayVertexAttribFormatEXT",
    "VertexArrayVertexAttribIFormatEXT",
    "VertexArrayVertexAttribLFormatEXT",
    "VertexArrayVertexAttribBindingEXT",
    "VertexArrayVertexBindingDivisorEXT",
    "NamedBufferPageCommitmentEXT",
    "NamedStringARB",
    "DeleteNamedStringARB",
    "CompileShaderIncludeARB",
    "IsNamedStringARB",
    "GetNamedStringARB",
    "GetNamedStringivARB",
    "EGLImageTargetTexStorageEXT" ,
    "EGLImageTargetTextureStorageEXT" ,
    "CopyImageSubDataNV",
    "ViewportSwizzleNV",
    "AlphaToCoverageDitherControlNV",
    "InternalBufferSubDataCopyMESA",
    "Vertex2hNV",
    "Vertex2hvNV",
    "Vertex3hNV",
    "Vertex3hvNV",
    "Vertex4hNV",
    "Vertex4hvNV",
    "Normal3hNV",
    "Normal3hvNV",
    "Color3hNV",
    "Color3hvNV",
    "Color4hNV",
    "Color4hvNV",
    "TexCoord1hNV",
    "TexCoord1hvNV",
    "TexCoord2hNV",
    "TexCoord2hvNV",
    "TexCoord3hNV",
    "TexCoord3hvNV",
    "TexCoord4hNV",
    "TexCoord4hvNV",
    "MultiTexCoord1hNV",
    "MultiTexCoord1hvNV",
    "MultiTexCoord2hNV",
    "MultiTexCoord2hvNV",
    "MultiTexCoord3hNV",
    "MultiTexCoord3hvNV",
    "MultiTexCoord4hNV",
    "MultiTexCoord4hvNV",
    "FogCoordhNV",
    "FogCoordhvNV",
    "SecondaryColor3hNV",
    "SecondaryColor3hvNV",
    "InternalSetError",
    "VertexAttrib1hNV",
    "VertexAttrib1hvNV",
    "VertexAttrib2hNV",
    "VertexAttrib2hvNV",
    "VertexAttrib3hNV",
    "VertexAttrib3hvNV",
    "VertexAttrib4hNV",
    "VertexAttrib4hvNV",
    "VertexAttribs1hvNV",
    "VertexAttribs2hvNV",
    "VertexAttribs3hvNV",
    "VertexAttribs4hvNV",
    "TexPageCommitmentARB",
    "TexturePageCommitmentEXT",
    "ImportMemoryWin32HandleEXT",
    "ImportSemaphoreWin32HandleEXT",
    "ImportMemoryWin32NameEXT",
    "ImportSemaphoreWin32NameEXT",
    "GetObjectLabelEXT",
    "LabelObjectEXT",
    "DrawArraysUserBuf",
    "DrawElementsUserBuf",
    "MultiDrawArraysUserBuf",
    "MultiDrawElementsUserBuf",
    "DrawArraysInstancedBaseInstanceDrawID",
    "DrawElementsInstancedBaseVertexBaseInstanceDrawID",
    "InternalInvalidateFramebufferAncillaryMESA",
    "InternalReleaseBufferMESA",
    "DrawElementsPacked",
    "DrawElementsUserBufPacked",
    "TexStorageAttribs2DEXT",
    "TexStorageAttribs3DEXT",
    "FramebufferTextureMultiviewOVR",
    "NamedFramebufferTextureMultiviewOVR",
    "FramebufferTextureMultisampleMultiviewOVR",
    "CreateSemaphoresNV",
    "GetSemaphoreParameterivNV",
    "SemaphoreParameterivNV",
    "DrawMeshTasksEXT",
    "DrawMeshTasksIndirectEXT",
    "MultiDrawMeshTasksIndirectEXT",
    "MultiDrawMeshTasksIndirectCountEXT",

    # Keep these last. They are never used by any app.
    "ColorTable",
    "ColorTableParameterfv",
    "ColorTableParameteriv",
    "CopyColorTable",
    "GetColorTable",
    "GetColorTableParameterfv",
    "GetColorTableParameteriv",
    "ColorSubTable",
    "CopyColorSubTable",
    "ConvolutionFilter1D",
    "ConvolutionFilter2D",
    "ConvolutionParameterf",
    "ConvolutionParameterfv",
    "ConvolutionParameteri",
    "ConvolutionParameteriv",
    "CopyConvolutionFilter1D",
    "CopyConvolutionFilter2D",
    "GetConvolutionFilter",
    "GetConvolutionParameterfv",
    "GetConvolutionParameteriv",
    "GetSeparableFilter",
    "SeparableFilter2D",
    "GetHistogram",
    "GetHistogramParameterfv",
    "GetHistogramParameteriv",
    "GetMinmax",
    "GetMinmaxParameterfv",
    "GetMinmaxParameteriv",
    "Histogram",
    "Minmax",
    "ResetHistogram",
    "ResetMinmax",
    "GetnColorTableARB",
    "GetnConvolutionFilterARB",
    "GetnHistogramARB",
    "GetnMinmaxARB",
    "GetnSeparableFilterARB",
]

offsets = {}
i = 0
for name in all_functions:
    if name in offsets:
        print('duplicated name in all_functions: ' + name)
        assert False
    offsets[name] = i
    i += 1

function_count = i

# Exported from libGL.so
def get_libgl_public_functions(libgl_symbols_filename):
    libgl_public_functions = set()
    with open(libgl_symbols_filename, 'r') as f:
        for line in f:
            if len(line) > 3 and line[0:2] == 'gl' and line[2] != 'X':
                libgl_public_functions.add(line[2:].rstrip())
    return libgl_public_functions



"""
A structured logging utility supporting multiple data formats such as CSV, JSON,
and YAML.

The main purpose of this script, besides having relevant information available
in a condensed and deserialized.

This script defines a protocol for different file handling strategies and provides
implementations for CSV, JSON, and YAML formats. The main class, StructuredLogger,
allows for easy interaction with log data, enabling users to load, save, increment,
set, and append fields in the log. The script also includes context managers for
file locking and editing log data to ensure data integrity and avoid race conditions.
"""

import json
import os
from collections.abc import MutableMapping, MutableSequence
from contextlib import contextmanager
from datetime import datetime
from pathlib import Path
from typing import Any, Protocol

import fire
from filelock import FileLock

try:
    import polars as pl

    CSV_LIB_EXCEPTION = None
except ImportError as e:
    CSV_LIB_EXCEPTION: ImportError = e

try:
    from ruamel.yaml import YAML

    YAML_LIB_EXCEPTION = None
except ImportError as e:
    YAML_LIB_EXCEPTION: ImportError = e


class ContainerProxy:
    """
    A proxy class that wraps a mutable container object (such as a dictionary or
    a list) and calls a provided save_callback function whenever the container
    or its contents
    are changed.
    """
    def __init__(self, container, save_callback):
        self.container = container
        self.save_callback = save_callback

    def __getitem__(self, key):
        value = self.container[key]
        if isinstance(value, (MutableMapping, MutableSequence)):
            return ContainerProxy(value, self.save_callback)
        return value

    def __setitem__(self, key, value):
        self.container[key] = value
        self.save_callback()

    def __delitem__(self, key):
        del self.container[key]
        self.save_callback()

    def __getattr__(self, name):
        attr = getattr(self.container, name)

        if callable(attr):
            def wrapper(*args, **kwargs):
                result = attr(*args, **kwargs)
                self.save_callback()
                return result

            return wrapper
        return attr

    def __iter__(self):
        return iter(self.container)

    def __len__(self):
        return len(self.container)

    def __repr__(self):
        return repr(self.container)


class AutoSaveDict(dict):
    """
    A subclass of the built-in dict class with additional functionality to
    automatically save changes to the dictionary. It maintains a timestamp of
    the last modification and automatically wraps nested mutable containers
    using ContainerProxy.
    """
    timestamp_key = "_timestamp"

    def __init__(self, *args, save_callback, register_timestamp=True, **kwargs):
        self.save_callback = save_callback
        self.__register_timestamp = register_timestamp
        self.__heartbeat()
        super().__init__(*args, **kwargs)
        self.__wrap_dictionaries()

    def __heartbeat(self):
        if self.__register_timestamp:
            self[AutoSaveDict.timestamp_key] = datetime.now().isoformat()

    def __save(self):
        self.__heartbeat()
        self.save_callback()

    def __wrap_dictionaries(self):
        for key, value in self.items():
            if isinstance(value, MutableMapping) and not isinstance(
                value, AutoSaveDict
            ):
                self[key] = AutoSaveDict(
                    value, save_callback=self.save_callback, register_timestamp=False
                )

    def __setitem__(self, key, value):
        if isinstance(value, MutableMapping) and not isinstance(value, AutoSaveDict):
            value = AutoSaveDict(
                value, save_callback=self.save_callback, register_timestamp=False
            )
        super().__setitem__(key, value)

        if self.__register_timestamp and key == AutoSaveDict.timestamp_key:
            return
        self.__save()

    def __getitem__(self, key):
        value = super().__getitem__(key)
        if isinstance(value, (MutableMapping, MutableSequence)):
            return ContainerProxy(value, self.__save)
        return value

    def __delitem__(self, key):
        super().__delitem__(key)
        self.__save()

    def pop(self, *args, **kwargs):
        result = super().pop(*args, **kwargs)
        self.__save()
        return result

    def update(self, *args, **kwargs):
        super().update(*args, **kwargs)
        self.__wrap_dictionaries()
        self.__save()


class StructuredLoggerStrategy(Protocol):
    def load_data(self, file_path: Path) -> dict:
        pass

    def save_data(self, file_path: Path, data: dict) -> None:
        pass


class CSVStrategy:
    def __init__(self) -> None:
        if CSV_LIB_EXCEPTION:
            raise RuntimeError(
                "Can't parse CSV files. Missing library"
            ) from CSV_LIB_EXCEPTION

    def load_data(self, file_path: Path) -> dict:
        dicts: list[dict[str, Any]] = pl.read_csv(
            file_path, try_parse_dates=True
        ).to_dicts()
        data = {}
        for d in dicts:
            for k, v in d.items():
                if k != AutoSaveDict.timestamp_key and k in data:
                    if isinstance(data[k], list):
                        data[k].append(v)
                        continue
                    data[k] = [data[k], v]
                else:
                    data[k] = v
        return data

    def save_data(self, file_path: Path, data: dict) -> None:
        pl.DataFrame(data).write_csv(file_path)


class JSONStrategy:
    def load_data(self, file_path: Path) -> dict:
        return json.loads(file_path.read_text())

    def save_data(self, file_path: Path, data: dict) -> None:
        with open(file_path, "w") as f:
            json.dump(data, f, indent=2)


class YAMLStrategy:
    def __init__(self):
        if YAML_LIB_EXCEPTION:
            raise RuntimeError(
                "Can't parse YAML files. Missing library"
            ) from YAML_LIB_EXCEPTION
        self.yaml = YAML()
        self.yaml.indent(sequence=4, offset=2)
        self.yaml.default_flow_style = False
        self.yaml.representer.add_representer(AutoSaveDict, self.represent_dict)

    @classmethod
    def represent_dict(cls, dumper, data):
        return dumper.represent_mapping("tag:yaml.org,2002:map", data)

    def load_data(self, file_path: Path) -> dict:
        return self.yaml.load(file_path.read_text())

    def save_data(self, file_path: Path, data: dict) -> None:
        with open(file_path, "w") as f:
            self.yaml.dump(data, f)


class StructuredLogger:
    def __init__(
        self, file_name: str, strategy: StructuredLoggerStrategy = None, truncate=False
    ):
        self.file_name: str = file_name
        self.file_path = Path(self.file_name)
        self._data: AutoSaveDict = AutoSaveDict(save_callback=self.save_data)

        if strategy is None:
            self.strategy: StructuredLoggerStrategy = self.guess_strategy_from_file(
                self.file_path
            )
        else:
            self.strategy = strategy

        if not self.file_path.exists():
            Path.mkdir(self.file_path.parent, exist_ok=True)
            self.save_data()
            return

        if truncate:
            with self.get_lock():
                os.truncate(self.file_path, 0)
                self.save_data()

    def load_data(self):
        self._data = self.strategy.load_data(self.file_path)

    def save_data(self):
        self.strategy.save_data(self.file_path, self._data)

    @property
    def data(self) -> AutoSaveDict:
        return self._data

    @contextmanager
    def get_lock(self):
        with FileLock(f"{self.file_path}.lock", timeout=10):
            yield

    @contextmanager
    def edit_context(self):
        """
        Context manager that ensures proper loading and saving of log data when
        performing multiple modifications.
        """
        with self.get_lock():
            try:
                self.load_data()
                yield
            finally:
                self.save_data()

    @staticmethod
    def guess_strategy_from_file(file_path: Path) -> StructuredLoggerStrategy:
        file_extension = file_path.suffix.lower().lstrip(".")
        return StructuredLogger.get_strategy(file_extension)

    @staticmethod
    def get_strategy(strategy_name: str) -> StructuredLoggerStrategy:
        strategies = {
            "csv": CSVStrategy,
            "json": JSONStrategy,
            "yaml": YAMLStrategy,
            "yml": YAMLStrategy,
        }

        try:
            return strategies[strategy_name]()
        except KeyError as e:
            raise ValueError(f"Unknown strategy for: {strategy_name}") from e


if __name__ == "__main__":
    fire.Fire(StructuredLogger)



#!/usr/bin/env python3
# Copyright © 2025 Collabora Ltd.
# Authors:
#   Sergi Blanch Torne <sergi.blanch.torne@collabora.com>
#
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from datetime import timedelta

from unittest import TestCase
from typing import Iterable

from marge_queue import MargeMergeRequest, MargeQueue, get_merge_queue


DATA = {
    37564: {
        # sample of a Merge Request rejected by marge and later reassigned to marge
        "updated_at": "2025-09-25T10:59:27.144Z",
        "title": "etnaviv: Fix util_blitter_save_so_targets(..) call",
        "notes": [
            # Those notes must be listed in reverse order
            # because the list is requested sort="desc"
            {
                "body": "assigned to @marge-bot and unassigned @austriancoder",
                "created_at": "2025-09-25T11:44:01.672Z",
            },
            {
                "body": "added 1 commit",
                "created_at": "2025-09-25T10:43:43.490Z",
            },
            {
                "body": "This branch couldn't be merged:",
                "created_at": "2025-09-25T10:42:27.817Z",
            },
            {
                "body": "assigned to @austriancoder and unassigned @marge-bot",
                "created_at": "2025-09-25T10:42:15.404Z",
            },
            {
                "body": "added 3 commits",
                "created_at": "2025-09-25T10:38:57.332Z",
            },
            {
                "body": "assigned to @marge-bot",
                "created_at": "2025-09-25T10:38:03.993Z",
            },
        ],
    },
    37560: {
        # sample of a Merge Request assigned to marge between the two
        # assignments of the previous sample. This is meant to show that
        # this goes first in the queue, and it doesn't confuses the sequence.
        "updated_at": "2025-09-25T11:38:08.420Z",
        "title": "tu: limit query pool types logged into RMV",
        "notes": [
            {
                "body": "assigned to @marge-bot",
                "created_at": "2025-09-25T10:47:40.040Z",
            },
            {
                "body": "added 1 commit",
                "created_at": "2025-09-25T10:46:42.113Z",
            },
            {
                "body": "mentioned in issue #13970",
                "created_at": "2025-09-25T10:45:35.566Z",
            },
            {
                "body": "Rb",
                "created_at": "2025-09-25T08:25:51.527Z",
            },
        ],
    },
}


@dataclass
class ProjectMergeRequestNote:
    body: str
    created_at: str


@dataclass
class ProjectMergeRequestNoteManager:
    """
    Mock the gitlab.v4.objects.ProjectMergeRequestNoteManager
    """
    iid: int

    def list(self, *args, **kwargs) -> Iterable:
        for note in DATA[self.iid]["notes"]:
            yield ProjectMergeRequestNote(**note)


@dataclass
class ProjectMergeRequest:
    """
    Mock the gitlab.v4.objects.ProjectMergeRequest
    """
    iid: int

    @property
    def updated_at(self) -> str:
        return DATA[self.iid]["updated_at"]

    @property
    def web_url(self) -> str:
        return f"https://gitlab.freedesktop.org/mesa/mesa/-/merge_requests/{self.iid}"

    @property
    def title(self) -> str:
        return DATA[self.iid]["title"]

    @property
    def notes(self) -> ProjectMergeRequestNoteManager:
        return ProjectMergeRequestNoteManager(self.iid)


@dataclass
class ProjectMergeRequestManager:
    """
    Mock the gitlab.v4.objects.ProjectMergeRequestManager
    """
    def list(self, *args, **kwargs) -> Iterable:
        for mr in DATA.keys():
            yield ProjectMergeRequest(mr)


@dataclass
class Project:
    """
    Mock the gitlab.v4.objects.Project
    """
    @property
    def mergerequests(self) -> ProjectMergeRequestManager:
        return ProjectMergeRequestManager()


class MargeMergeTestCase(TestCase):
    def test_queue(self) -> None:
        queue = MargeQueue()
        for mr_id in DATA.keys():
            gl_mr_mock_obj = ProjectMergeRequest(mr_id)
            marge_mr = MargeMergeRequest(gl_mr_mock_obj)

            # for each element in the data sample, test the MargeMergeRequest construction
            self.assertEqual(marge_mr.title, gl_mr_mock_obj.title)
            self.assertEqual(marge_mr.web_url, gl_mr_mock_obj.web_url)
            self.assertIsInstance(marge_mr.time_enqueued, timedelta)

            queue.append(marge_mr)

        self.assertEqual(queue.n_merge_requests_enqueued, 2)


    def test_get_merge_queue(self) -> None:
        project = Project()
        queue = get_merge_queue(project)

        self.assertEqual(queue.n_merge_requests_enqueued, 2)
        self.assertEqual(len(queue.sorted_queue), 2)
        # They are in reverse order because of the reassignment to marge
        self.assertEqual(queue.sorted_queue[0].id, list(DATA.keys())[1])
        self.assertEqual(queue.sorted_queue[1].id, list(DATA.keys())[0])
        self.assertEqual(len(queue.undetermined), 0)

        for marge_mr in queue.sorted_queue:
            mr_id = marge_mr.id
            self.assertEqual(marge_mr.title, DATA[mr_id]["title"])
            self.assertEqual(
                marge_mr.web_url,
                f"https://gitlab.freedesktop.org/mesa/mesa/-/merge_requests/{mr_id}"
            )
            self.assertIsInstance(marge_mr.time_enqueued, timedelta)
